# bleak_wrapper.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue Sep  7 21:35:53 UTC 2021

"""Manage a Bluetooth connection using bleak."""

import asyncio
import logging
import threading
from typing import Pattern, Dict, Any, Callable, Optional

from bleak import BleakScanner, BleakClient
from bleak.backends.device import BLEDevice as BleakDevice

from open_gopro.exceptions import ConnectFailed
from open_gopro.util import Singleton
from open_gopro.ble import (
    Service,
    Characteristic,
    Descriptor,
    AttributeTable,
    BLEController,
    NotiHandlerType,
    FailedToFindDevice,
    UUID,
)

logger = logging.getLogger(__name__)


class BleakWrapperController(BLEController[BleakDevice, BleakClient], Singleton):
    """Wrapper around bleak to manage a Bluetooth connection.

    Note, this is a singleton.
    """

    def __init__(self) -> None:
        # Thread to run ble controller asyncio loop (to abstract asyncio from client as well as handle async notifications)
        self._module_loop: asyncio.AbstractEventLoop  # Will be set when module thread starts
        self._module_thread = threading.Thread(daemon=True, target=self._run, name="data")
        self._ready = threading.Event()
        self._module_thread.start()
        self._ready.wait()

    def _run(self) -> None:
        """Thread to keep the event loop running to interface with the BLE adapter."""
        # Create loop for this new thread
        self._module_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._module_loop)

        # Run forever
        self._ready.set()
        self._module_loop.run_forever()

    def _as_coroutine(self, action: Callable, timeout: float = None) -> Any:
        """Run a function as a coroutine in the module thread.

        This will transfer execution of the given partial to the module thread of this instance

        Args:
            action (Callable): function and parameters to run as corouting
            timeout (float, optional): Time to wait for coroutine to return (in seconds). Defaults to None (wait forever).

        Returns:
            Any: Passes return of coroutine through
        """
        # Allow timeout exception to propagate
        return asyncio.run_coroutine_threadsafe(action(), self._module_loop).result(timeout)

    def read(self, handle: BleakClient, uuid: str) -> bytearray:
        """read data from a UUID.

        Args:
            handle (BleakClient): client to read from
            uuid (str): uuid to read

        Returns:
            bytearray: read data
        """

        async def _async_read() -> bytearray:  # pylint: disable=missing-return-doc
            logger.debug(f"Reading from {uuid}")
            response = await handle.read_gatt_char(uuid)
            logger.debug(f'Received response on {uuid}: {response.hex( ":")}')
            return response

        return self._as_coroutine(_async_read)

    def write(self, handle: BleakClient, uuid: str, data: bytearray) -> None:
        """Write data to a UUID.

        Perform a write, then wait for event to be notified (from a notification handler)

        Args:
            handle (BleakClient): Device to write to
            uuid (str): characteristic UUID to write to
            data (bytearray): data to write
        """

        async def _async_write() -> None:
            logger.debug(f"Writing to {uuid}: {data.hex(':')}")
            await handle.write_gatt_char(uuid, data)

        self._as_coroutine(_async_write)

    def scan(self, token: Pattern, timeout: int = 5) -> BleakDevice:
        """scan for a regex.

        Args:
            token (Pattern): Regex to look for when scanning.
            timeout (int, optional): Time to scan. Defaults to 5.

        Raises:
            FailedToFindDevice: Did not find any of the token when scanning.

        Returns:
            BleakDevice: The first matched device that was discovered
        """

        async def _async_scan() -> BleakDevice:  # pylint: disable=missing-return-doc
            logger.info(f"Scanning for {token.pattern} bluetooth devices...")
            devices: Dict[str, BleakDevice] = {}

            def _scan_callback(device: BleakDevice, _: Any) -> None:
                """Bleak optional scan callback to receive every scan result.

                We need this because the GoPro will sometimes only show as a
                nonconnectable scan response from bleak.

                Also, if the device shows as both the above as well as a connectable advertisement,
                the latter must be used in order to avoid an exception from CoreBluetooth on Mac.

                Args:
                    device (BleakDevice): discovered device
                    _ : advertisement that we're ignoring
                """
                # Add to the dict if not unknown
                if device.name != "Unknown" and device.name is not None:
                    devices[device.name] = device

            # Now get list of connectable advertisements
            for device in await BleakScanner.discover(timeout=timeout, detection_callback=_scan_callback):
                if device.name != "Unknown" and device.name is not None:
                    devices[device.name] = device
            for d in devices:
                logger.info(f"\tDiscovered: {d}")

            # Now look for our matching device(s)
            matched_devices = [device for name, device in devices.items() if token.match(name)]
            if len(matched_devices) == 0:
                raise FailedToFindDevice
            logger.info(f"Found {len(matched_devices)} matching devices.")
            # If there's more than 1, the first one gets lucky.
            return matched_devices[0]

        return self._as_coroutine(_async_scan)

    def connect(
        self,
        disconnect_cb: Callable,
        device: BleakDevice,
        timeout: int = 15,
    ) -> BleakClient:
        """Connect to a device.

        Args:
            disconnect_cb (Callable): function called when a disconnect is received
            device (BleakDevice): Device to connect to
            timeout (int, optional): How long to try connecting before timing out and raising exception. Defaults to 15.

        Raises:
            ConnectFailed: Connection failed to establish

        Returns:
            BleakClient: Connected device
        """

        async def _async_connect() -> Optional[BleakClient]:  # pylint: disable=missing-return-doc
            logger.info(f"Establishing BLE connection to {device}...")
            bleak_client = BleakClient(device, disconnected_callback=disconnect_cb)
            try:
                await bleak_client.connect(timeout=timeout)
            except Exception as e:  # pylint: disable= broad-except
                logger.warning(f"Bleak failed to connect: {repr(e)}")
                return None

            return bleak_client

        client = self._as_coroutine(_async_connect)
        if client is None:
            raise ConnectFailed("Bleak Connection Timed out", timeout, 1)
        return client

    def pair(self, handle: BleakClient) -> None:
        """pair to a device after connection.

        This is required for Windows and not allowed on Mac...

        Args:
            handle (BleakClient): Device to pair to
        """

        async def _async_def_pair() -> None:
            logger.debug("Attempting to pair...")
            try:
                await handle.pair()
            except NotImplementedError:
                # This is expected on Mac
                pass
            logger.debug("Pairing complete!")

        self._as_coroutine(_async_def_pair)

    def enable_notifications(self, handle: BleakClient, handler: NotiHandlerType) -> None:
        """enable all notifications.

        Search through all characteristics and enable any that have notification property.

        Args:
            handle (BleakClient): Device to enable notifications for
            handler (Callable): Notification callback handler
        """

        async def _async_enable_notifications() -> None:
            logger.info("Enabling notifications...")
            for service in handle.services:
                for char in service.characteristics:
                    if "notify" in char.properties:
                        logger.debug(f"Enabling notification on char {char.uuid}")
                        await handle.start_notify(char, handler)
            logger.info("Done enabling notifications")

        self._as_coroutine(_async_enable_notifications)

    def discover_chars(self, handle: BleakClient) -> AttributeTable:
        """discover all characteristics.

        Args:
            handle (BleakClient): Device to perform discovery on

        Returns:
            AttributeTable: Dict of services by UUID
        """

        async def _async_discover_chars() -> AttributeTable:  # pylint: disable=missing-return-doc
            logger.info("Discovering characteristics...")
            services: Dict[UUID, Service] = {}

            for service in handle.services:
                logger.debug("[Service] {0}: {1}".format(service.uuid, service.description))
                try:
                    # Create new service
                    services[UUID(service.uuid)] = Service(UUID(service.uuid), service.description)
                except ValueError:
                    logger.error(f"{service.uuid} is not a known service")
                    continue

                # Loop over all chars in service
                chars: Dict[UUID, Characteristic] = {}
                for char in service.characteristics:
                    # Read if applicable
                    value = bytes(await handle.read_gatt_char(char.uuid)) if "read" in char.properties else b""
                    # Create and log characteristic
                    c = Characteristic(char.handle, UUID(char.uuid), char.properties, char.description, value)
                    # Get any descriptors if they exist
                    descriptors = []
                    for descriptor in char.descriptors:
                        value = await handle.read_gatt_descriptor(descriptor.handle)
                        descriptors.append(Descriptor(descriptor.handle, value))
                    # Update char and add to dict
                    c.descriptors = descriptors
                    # Add characteristic to char dic
                    chars[c.uuid] = c

                # Add char dict to service
                services[UUID(service.uuid)].chars = chars

            logger.info("Done discovering characteristics!")
            return AttributeTable(services)

        return self._as_coroutine(_async_discover_chars)

    def disconnect(self, handle: BleakClient) -> None:
        """terminate a BLE connection.

        Args:
            handle (BleakClient): client to disconnect from

        Returns:
            bool: True if disconnect was successful, False otherwise
        """

        async def _async_disconnect() -> None:
            if handle.is_connected:
                logger.info("Disconnecting...")
                await handle.disconnect()
                # Disconnect handler registered during connect will be asynchronously called
            logger.info("Device disconnected!")

        return self._as_coroutine(_async_disconnect)



# communication_client.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue Sep  7 21:35:53 UTC 2021

"""GoPro specific BLE client"""

import re
import logging
from pathlib import Path
from abc import ABC, abstractmethod
from typing import Generic, Union, Pattern, Optional

from open_gopro.ble import (
    BleDevice,
    BleHandle,
    BLEController,
    DisconnectHandlerType,
    NotiHandlerType,
    BleClient,
    UUID,
)
from open_gopro.wifi import WifiClient, WifiController
from open_gopro.responses import GoProResp, ParserMapType, Parser
from open_gopro.constants import ProducerType, ResponseType

logger = logging.getLogger(__name__)


class GoProResponder:
    """Interface implementation for accumulating response GoPro response parsers"""

    def __init__(self) -> None:
        # Will be filled out by Command, Setting, Status class's when instantiated
        self._response_parsers: ParserMapType = {}

    @property
    def _parser_map(self) -> ParserMapType:
        return self._response_parsers

    def _add_parser(self, identifier: ResponseType, parser: Parser) -> None:
        self._response_parsers[identifier] = parser


class GoProWifi(ABC, GoProResponder):
    """GoPro specific WiFi Client

    Args:
        controller (WifiController): instance of Wifi Controller to use for this client
    """

    def __init__(self, controller: WifiController):
        GoProResponder.__init__(self)
        self._wifi: WifiClient = WifiClient(controller)

    @abstractmethod
    def _get(self, url: str) -> GoProResp:
        """Send an HTTP GET request to a string endpoint.

        Args:
            url (str): endpoint not including GoPro base path

        Returns:
            GoProResp: GoPro response
        """
        raise NotImplementedError

    @abstractmethod
    def _stream_to_file(self, url: str, file: Path) -> None:
        """Send an HTTP GET request to an Open GoPro endpoint to download a binary file.

        Args:
            url (str): endpoint URL
            file (Path): location where file should be downloaded to
        """
        raise NotImplementedError

    @property
    def password(self) -> Optional[str]:
        """Get the GoPro AP's password

        Returns:
            Optional[str]: password or None if it is not known
        """
        return self._wifi.password

    @property
    def ssid(self) -> Optional[str]:
        """Get the GoPro AP's WiFi SSID

        Returns:
            Optional[str]: SSID or None if it is not known
        """
        return self._wifi.ssid


class GoProBle(ABC, GoProResponder, Generic[BleHandle, BleDevice]):
    """GoPro specific BLE Client

    Args:
        controller (BleController): instance of BLE Controller to use for this client
    """

    def __init__(
        self,
        controller: BLEController,
        disconnected_cb: DisconnectHandlerType,
        notification_cb: NotiHandlerType,
        target: Optional[Union[Pattern, BleDevice]],
    ) -> None:
        GoProResponder.__init__(self)
        self._ble: BleClient = BleClient(
            controller,
            disconnected_cb,
            notification_cb,
            re.compile(r"GoPro [A-Z0-9]{4}") if target is None else target,
        )

    @abstractmethod
    def _register_listener(self, producer: ProducerType) -> None:
        """Register to receive notifications for a producer.

        Args:
            producer (ProducerType): producer that we want to receive notifications from
        """
        raise NotImplementedError

    @abstractmethod
    def _unregister_listener(self, producer: ProducerType) -> None:
        """Stop receiving notifications from a producer.

        Args:
            producer (ProducerType): Producer to stop receiving notifications for
        """
        raise NotImplementedError

    @abstractmethod
    def get_update(self, timeout: float = None) -> GoProResp:
        """Get a notification that was received from a registered producer.

        Args:
            timeout (float, optional): Time to wait for a notification before returning. Defaults to None (wait forever)

        Returns:
            GoProResp: the received update
        """
        raise NotImplementedError

    @abstractmethod
    def _write_characteristic_receive_notification(self, uuid: UUID, data: bytearray) -> GoProResp:
        raise NotImplementedError

    @abstractmethod
    def _read_characteristic(self, uuid: UUID) -> GoProResp:
        raise NotImplementedError



# demo.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Wed, Sep  1, 2021  5:05:45 PM

"""Demonstration of using the GoPro package."""

import time
import logging
import argparse
import threading
from pathlib import Path
from typing import Tuple, Optional

from rich.console import Console

from open_gopro import GoPro
from open_gopro.constants import QueryCmdId
from open_gopro.util import launch_vlc, setup_logging, set_logging_level

logger = logging.getLogger(__name__)
console = Console()  # rich consoler printer


def process_async_events(gopro: GoPro) -> None:
    """Separate thread to continuously get and print asynchronous events from the GoPro instance.

    Args:
        gopro (GoPro): instance to get updates from
    """
    while True:
        # Block until we receive one
        update = gopro.get_update()

        # Print pushed updates (they're either setting capability or setting / status value updates)
        description = "capabilities are" if update.cmd is QueryCmdId.SETTING_CAPABILITY_PUSH else "value is"
        for item in update:
            console.print(f"{item} {description} now: {update[item]}")


def main() -> int:
    """Main demo functionality

    Returns:
        int: error code
    """
    identifier, log_location, vlc_location = parse_arguments()
    global logger
    logger = setup_logging(logger, log_location)

    gopro: Optional[GoPro] = None
    return_code = 0
    try:
        with GoPro(identifier) as gopro:
            # Dump services to CSV
            gopro._ble.services_as_csv()

            # Start a thread to handle asynchronous notifications
            threading.Thread(target=process_async_events, args=(gopro,), daemon=True).start()

            # Now we only want errors and above since we're going to be doing a lot of printing below
            set_logging_level(logger, logging.ERROR)

            with console.status("[bold green]Ensuring the shutter is off..."):
                if gopro.is_encoding:
                    assert gopro.ble_command.set_shutter(gopro.params.Shutter.OFF).is_ok

            with console.status("[bold green]Starting with known state..."):
                assert gopro.ble_command.set_turbo_mode(False).is_ok
                # Ensure we are in a mode that can enter photo preset
                if gopro.version >= 2.0:
                    assert gopro.ble_setting.video_performance_mode.set(
                        gopro.params.VideoPerformanceMode.MAX_PERFORMANCE
                    ).is_ok
                    assert gopro.ble_setting.max_lens_mode.set(gopro.params.MaxLensMode.DEFAULT).is_ok

            with console.status("[bold green]Getting all statuses via BLE..."):
                for status, value in gopro.ble_command.get_camera_statuses().items():
                    console.print(f"{status} : {value}")

            with console.status("[bold green]Getting settings and statuses via Wifi..."):
                for item, value in gopro.wifi_command.get_camera_state().items():
                    console.print(f"{item} : {value}")

            with console.status("[bold green]Getting Open GoPro info..."):
                assert gopro.keep_alive()
                assert gopro.ble_command.set_third_party_client_info().is_ok
                version = gopro.ble_command.get_open_gopro_api_version()
                assert version.is_ok
                console.print(f"Version is {version['major']}.{version['minor']}")

            with console.status("[bold green]Setting Cinematic preset..."):
                assert gopro.ble_command.load_preset(gopro.params.Preset.CINEMATIC).is_ok

            with console.status("[bold green]Getting some individual values via BLE..."):
                resolution = gopro.ble_setting.resolution.get_value().flatten
                console.print(f"Current resolution is {resolution}")
                fov = gopro.ble_setting.video_field_of_view.get_value().flatten
                console.print(f"Current video FOV is {fov}")
                encoding = gopro.ble_status.encoding_active.get_value().flatten
                console.print(f"Current encoding status is {encoding}")

            with console.status("[bold green]Getting some capabilities via BLE..."):
                resolution_caps = gopro.ble_setting.resolution.get_capabilities_values().flatten
                console.print(f"Current resolution capabilities are: {resolution_caps}")
                fov_caps = gopro.ble_setting.video_field_of_view.get_capabilities_values().flatten
                console.print(f"Current video FOV capabilities are: {fov_caps}")

            with console.status("[bold green]Registering for push notifications..."):
                assert gopro.ble_setting.resolution.set(gopro.params.Resolution.RES_1080).is_ok
                fps_caps_update = gopro.ble_setting.fps.register_value_update().flatten
                console.print(f"New fps capabilities are: {fps_caps_update}")
                assert gopro.ble_setting.resolution.register_value_update().is_ok
                assert gopro.ble_setting.resolution.register_capability_update().is_ok
                assert gopro.ble_setting.fps.register_capability_update().is_ok
                assert gopro.ble_setting.video_field_of_view.register_value_update().is_ok
                assert gopro.ble_setting.video_field_of_view.register_capability_update().is_ok
                assert gopro.ble_status.encoding_active.register_value_update().is_ok

            with console.status("[bold green]Taking a photo..."):
                assert gopro.ble_command.load_preset(gopro.params.Preset.PHOTO).is_ok
                assert gopro.ble_command.set_shutter(gopro.params.Shutter.ON).is_ok

            with console.status("[bold green]Taking a video..."):
                assert gopro.ble_command.load_preset(gopro.params.Preset.CINEMATIC).is_ok
                assert gopro.ble_command.set_shutter(gopro.params.Shutter.ON).is_ok
                time.sleep(2)  # Take a 2 second video
                assert gopro.ble_command.set_shutter(gopro.params.Shutter.OFF).is_ok

            with console.status("[bold green]Testing some capability pushes"):
                assert gopro.ble_setting.resolution.set(gopro.params.Resolution.RES_1080).is_ok
                assert gopro.ble_setting.fps.set(gopro.params.FPS.FPS_30).is_ok
                assert gopro.ble_setting.fps.set(gopro.params.FPS.FPS_240).is_ok
                assert gopro.ble_setting.resolution.get_capabilities_values().is_ok
                assert gopro.ble_setting.video_field_of_view.set(gopro.params.VideoFOV.LINEAR).is_ok
                assert gopro.ble_setting.video_field_of_view.set(gopro.params.VideoFOV.NARROW).is_ok
                assert gopro.ble_setting.video_field_of_view.set(gopro.params.VideoFOV.WIDE).is_ok
                # We expect this to fail as it is not valid with the current resolution
                assert not gopro.ble_setting.video_field_of_view.set(gopro.params.VideoFOV.MAX_SUPERVIEW).is_ok
                # Cycle through resolutions. This will cause other settings (and their capabilities) to update
                for resolution in resolution_caps:
                    assert gopro.ble_setting.resolution.set(resolution).is_ok

            with console.status("[bold green]Get media list and download some media..."):
                assert gopro.wifi_command.set_preset(gopro.params.Preset.PHOTO).is_ok
                media_list = gopro.wifi_command.get_media_list()["media"][0]["fs"]
                # Find a picture and download it
                picture = ""
                for file in [x["n"] for x in media_list]:
                    if file.lower().endswith(".jpg"):
                        picture = file
                        gopro.wifi_command.download_file(camera_file=picture)
                        break
                # Find a video and download it
                video = ""
                for file in [x["n"] for x in media_list]:
                    if file.lower().endswith(".mp4"):
                        video = file
                        gopro.wifi_command.download_file(camera_file=video)
                        break

            with console.status("[bold green]Getting media information..."):
                assert gopro.wifi_command.get_media_info(video).is_ok
                assert gopro.wifi_command.get_media_info(picture).is_ok
                gopro.wifi_command.get_gpmf_data(camera_file=picture)
                gopro.wifi_command.get_screennail(camera_file=video)
                gopro.wifi_command.get_telemetry(camera_file=video)
                gopro.wifi_command.get_thumbnail(camera_file=picture)
                assert gopro.wifi_command.get_preset_status().is_ok

            with console.status("[bold green]Cycling through resolutions via Wifi..."):
                assert gopro.wifi_command.set_preset(gopro.params.Preset.CINEMATIC).is_ok
                for resolution in gopro.params.Resolution:
                    assert gopro.wifi_setting.resolution.set(resolution).is_ok

            with console.status("[bold green]Starting the preview stream..."):
                gopro.wifi_command.stop_preview_stream()
                gopro.wifi_command.start_preview_stream()

            console.print("Preview stream is enabled. View with a UDP server at udp://@:8554.")
            console.print("Send keyboard interrupt to exit.")

            threading.Thread(target=launch_vlc, args=(vlc_location,), daemon=True).start()

            while True:
                time.sleep(0.2)

    except Exception as e:  # pylint: disable=broad-except
        logger.error(repr(e))
        return_code = 1
    except KeyboardInterrupt:
        logger.warning("Received keyboard interrupt. Shutting down...")
    finally:
        if gopro is not None:
            gopro.close()
        console.print("Exiting...")
        return return_code  # pylint: disable=lost-exception


def parse_arguments() -> Tuple[str, Path, Path]:
    """Parse command line arguments

    Returns:
        Tuple[str, Path, Path]: (identifier, path to save log, path to VLC)
    """
    parser = argparse.ArgumentParser(
        description="Connect to a GoPro camera via BLE and Wifi and do some things."
    )
    parser.add_argument(
        "-i",
        "--identifier",
        type=str,
        help="Last 4 digits of GoPro serial number, which is the last 4 digits of the default camera SSID. \
            If not used, first discovered GoPro will be connected to",
        default=None,
    )
    parser.add_argument(
        "-l",
        "--log",
        type=Path,
        help="Location to store detailed log",
        default="gopro_demo.log",
    )
    parser.add_argument(
        "-v",
        "--vlc",
        type=Path,
        help="VLC location. If not set, the location will attempt to be automatically discovered.",
        default=None,
    )
    args = parser.parse_args()

    return args.identifier, args.log, args.vlc


if __name__ == "__main__":
    main()



# services.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Wed, Sep  1, 2021  5:05:50 PM

"""Objects to nicely interact with BLE services, characteristics, and attributes."""

import json
import enum
from dataclasses import dataclass, field, asdict
from typing import Dict, List, Union


GOPRO_BASE_UUID = "b5f9{}-aa8d-11e3-9046-0002a5d5c51b"

# TODO this is temporary. Really UUID should be a data structure and we pass in the values
# of these UUID's from the application
class UUID(enum.Enum):
    """BLE UUID."""

    # Generic Attribute Service
    S_GENERIC_ATT = "00001801-0000-1000-8000-00805f9b34fb"

    # Generic Access Service
    S_GENERIC_ACCESS = "00001800-0000-1000-8000-00805f9b34fb"
    ACC_DEVICE_NAME = "00002a00-0000-1000-8000-00805f9b34fb"
    ACC_APPEARANCE = "00002a01-0000-1000-8000-00805f9b34fb"
    ACC_PREF_CONN_PARAMS = "00002a04-0000-1000-8000-00805f9b34fb"
    ACC_CENTRAL_ADDR_RES = "00002aa6-0000-1000-8000-00805f9b34fb"

    # Tx Power
    S_TX_POWER = "00001804-0000-1000-8000-00805f9b34fb"
    TX_POWER_LEVEL = "00002a07-0000-1000-8000-00805f9b34fb"

    # Battery Service
    S_BATTERY = "0000180f-0000-1000-8000-00805f9b34fb"
    BATT_LEVEL = "00002a19-0000-1000-8000-00805f9b34fb"

    # Device Information Service
    S_DEV_INFO = "0000180a-0000-1000-8000-00805f9b34fb"
    INF_MAN_NAME = "00002a29-0000-1000-8000-00805f9b34fb"
    INF_MODEL_NUM = "00002a24-0000-1000-8000-00805f9b34fb"
    INF_SERIAL_NUM = "00002a25-0000-1000-8000-00805f9b34fb"
    INF_FW_REV = "00002a26-0000-1000-8000-00805f9b34fb"
    INF_HW_REV = "00002a27-0000-1000-8000-00805f9b34fb"
    INF_SW_REV = "00002a28-0000-1000-8000-00805f9b34fb"
    INF_SYS_ID = "00002a23-0000-1000-8000-00805f9b34fb"
    INF_CERT_DATA = "00002a2a-0000-1000-8000-00805f9b34fb"
    INF_PNP_ID = "00002a50-0000-1000-8000-00805f9b34fb"

    # GoPro Wifi Access Point Service
    S_WIFI_ACCESS_POINT = GOPRO_BASE_UUID.format("0001")
    WAP_SSID = GOPRO_BASE_UUID.format("0002")
    WAP_PASSWORD = GOPRO_BASE_UUID.format("0003")
    WAP_POWER = GOPRO_BASE_UUID.format("0004")
    WAP_STATE = GOPRO_BASE_UUID.format("0005")
    WAP_CSI_PASSWORD = GOPRO_BASE_UUID.format("0006")

    # GoPro Control & Query Service
    S_CONTROL_QUERY = "0000fea6-0000-1000-8000-00805f9b34fb"
    CQ_COMMAND = GOPRO_BASE_UUID.format("0072")
    CQ_COMMAND_RESP = GOPRO_BASE_UUID.format("0073")
    CQ_SETTINGS = GOPRO_BASE_UUID.format("0074")
    CQ_SETTINGS_RESP = GOPRO_BASE_UUID.format("0075")
    CQ_QUERY = GOPRO_BASE_UUID.format("0076")
    CQ_QUERY_RESP = GOPRO_BASE_UUID.format("0077")
    CQ_SENSOR = GOPRO_BASE_UUID.format("0078")
    CQ_SENSOR_RESP = GOPRO_BASE_UUID.format("0079")

    # GoPro Camera Management Service
    S_CAMERA_MANAGEMENT = GOPRO_BASE_UUID.format("0090")
    CM_NET_MGMT_COMM = GOPRO_BASE_UUID.format("0091")
    CN_NET_MGMT_RESP = GOPRO_BASE_UUID.format("0092")

    # Unknown
    S_UNKNOWN = GOPRO_BASE_UUID.format("0080")
    INTERNAL_81 = GOPRO_BASE_UUID.format("0081")
    INTERNAL_82 = GOPRO_BASE_UUID.format("0082")
    INTERNAL_83 = GOPRO_BASE_UUID.format("0083")
    INTERNAL_84 = GOPRO_BASE_UUID.format("0084")


def get_gopro_desc(uuid: str) -> Union[UUID, str]:
    """Attempt to retrieve a the name of a UUID from it's value.

    Args:
        uuid (str): string representation of UUID

    Returns:
        Union[UUID, str]: a UUID object if success, otherwise just the input string
    """
    try:
        return UUID(uuid.lower()).name
    except ValueError:
        return uuid


@dataclass
class Descriptor:
    """A charactersistic descriptor.

    Args:
        handle (int) : the handle of the attribute table that the descriptor resides at
        value (bytes) : the byte stream value of the descriptor
    """

    handle: int
    value: bytes

    def __str__(self) -> str:  # pylint: disable=missing-return-doc
        return json.dumps(asdict(self), indent=4, default=str)


@dataclass
class Characteristic:
    """A BLE charactersistic.

    Args:
        handle (int) : the handle of the attribute table that the characteristic resides at
        uuid (UUID) : the UUID of the characteristic
        props (List[str]) : the characteristic's properties (READ, WRITE, NOTIFY, etc)
        name (str) : the characteristic's name
        value (bytes) : the byte stream value of the characteristic value
        descriptors (List[Descriptor], optional) : Any relevant descriptors if they exist
    """

    handle: int
    uuid: UUID
    props: List[str]
    name: str
    value: bytes
    descriptors: List[Descriptor] = field(default_factory=list)

    def __str__(self) -> str:  # pylint: disable=missing-return-doc
        return json.dumps(asdict(self), indent=4, default=str)


@dataclass
class Service:
    """A BLE service or grouping of Characteristics.

    Args:
        uuid (UUID) : the service's UUID
        name (str) : the service's name
        chars (Dict[str, Characteristic]) : the dictionary of characteristics, indexed by name
    """

    uuid: UUID
    name: str
    chars: Dict[UUID, Characteristic] = field(default_factory=dict)


class AttributeTable:
    """The attribute table to store / look up BLE services, characteristics, and attributes.

    Args:
        services (Dict[UUID, Service]): A dictionary of Services indexed by UUID..
    """

    def __init__(self, services: Dict[UUID, Service]) -> None:
        self.services = services

    def handle2uuid(self, handle: int) -> UUID:
        """Get a UUID from a handle.

        Args:
            handle (int): the handle to search for

        Raises:
            Exception: No characteristic was found at this handle

        Returns:
            UUID: The found UUID
        """
        for s in self.services.values():
            for c in s.chars.values():
                if c.handle == handle:
                    return c.uuid
        raise Exception(f"Matching UUID not found for handle {handle}")



# client.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue Sep  7 21:35:53 UTC 2021

"""Generic BLE Client definition that is composed of a BLE Controller."""

import re
import csv
import logging
from pathlib import Path
from typing import Generic, Optional, Union, Pattern

from open_gopro.exceptions import FailedToFindDevice, ConnectFailed
from .controller import (
    BLEController,
    BleDevice,
    BleHandle,
    DisconnectHandlerType,
    NotiHandlerType,
)
from .services import AttributeTable

logger = logging.getLogger(__name__)


class BleClient(Generic[BleHandle, BleDevice]):
    """A BLE client that is composed of, among other things, a BLE interface

    The interface is generic and can be set with the 'controller' argument
    """

    def __init__(
        self,
        controller: BLEController,
        disconnected_cb: DisconnectHandlerType,
        notification_cb: NotiHandlerType,
        target: Union[Pattern, BleDevice],
    ) -> None:
        if target is None:
            raise ValueError("Target can not be None!")
        if isinstance(target, str):
            # TODO is there a way to check if target is BleDevice?
            target = re.compile(target)
        self._controller = controller
        self._disconnected_cb = disconnected_cb
        self._notification_cb = notification_cb
        self._target = target
        self._gatt_table: Optional[AttributeTable] = None
        self._device: Optional[BleDevice] = None
        self._handle: Optional[BleHandle] = None
        self._identifier = None if isinstance(target, Pattern) else str(target)

    def __del__(self) -> None:
        logger.debug("In destructor...")
        self.close()

    def _find_device(self, timeout: int = 5, retries: int = 30) -> None:
        self._device = None
        assert isinstance(self._target, Pattern)
        for retry in range(1, retries):
            try:
                self._device = self._controller.scan(self._target, timeout)
                return
            except FailedToFindDevice as e:
                logger.warning(f"Failed to find a device in {timeout} seconds. Retrying #{retry}")
                if retry == retries - 1:
                    raise FailedToFindDevice from e

    def open(self, timeout: int = 10, retries: int = 5) -> None:
        """Open the client resource so that it is ready to send and receive data.

        Args:
            timeout (int, optional): How long to try connecting (in seconds) before retrying. Defaults to 10.
            retries(int, optional): How many retries to attempt before giving up. Defaults to 5
        """
        # If we need we need to find the device to connect
        if isinstance(self._target, Pattern):
            self._find_device()
        # Otherwise we already have it
        else:
            self._device = self._target
        self._identifier = str(self._device)

        logger.info("Establishing the BLE connection")
        for retry in range(retries):
            try:
                self._handle = self._controller.connect(self._disconnected_cb, self._device, timeout=timeout)
                break
            except ConnectFailed as e:
                logger.warning(f"Failed to connect in {timeout} seconds. Retrying #{retry}")
                if retry == retries - 1:
                    raise ConnectFailed("BLE", timeout, retries) from e

        assert self._handle is not None
        # Attempt to pair
        self._controller.pair(self._handle)
        # Discover characteristics
        self._gatt_table = self._controller.discover_chars(self._handle)
        # Enable all GATT notifications
        self._controller.enable_notifications(self._handle, self._notification_cb)

    def close(self) -> None:
        """Close the client resource.

        This should always be called before exiting.
        """
        if self.is_connected:
            logger.info("Terminating the BLE connection")
            self._controller.disconnect(self._handle)
            self._handle = None
        else:
            logger.debug("BLE already disconnected")

    def read(self, uuid: str) -> bytearray:
        """Read byte data from a characteristic (identified by UUID)

        Args:
            uuid (str): characteristic to read

        Returns:
            bytearray: byte data that was read
        """
        return self._controller.read(self._handle, uuid)

    def write(self, uuid: str, data: bytearray) -> None:
        """Write byte data to a characteristic (identified by UUID)

        Args:
            uuid (str): characteristic to write to
            data (bytearray): byte data to write
        """
        self._controller.write(self._handle, uuid, data)

    @property
    def gatt_table(self) -> AttributeTable:
        """Return the attribute table

        Returns:
            AttributeTable: table of BLE attributes
        """
        if self._gatt_table is None:
            # Discover characteristics
            assert self._handle is not None
            self._gatt_table = self._controller.discover_chars(self._handle)
        return self._gatt_table

    @property
    def identifier(self) -> Optional[str]:
        """A string that identifies the GoPro

        Returns:
            Optional[str]: identifier or None if client has not yet been discovered
        """
        return self._identifier

    @property
    def is_connected(self) -> bool:
        """Is BLE currently connected?

        Returns:
            bool: True if yes, False if no
        """
        return self._handle is not None

    @property
    def is_discovered(self) -> bool:
        """Has the target been discovered (via BLE scanning)?

        Returns:
            bool: True if yes, False if no
        """
        return self._device is not None

    def services_as_csv(self, dump_file: Path = Path("services.csv")) -> None:
        """Dump the services as a .csv

        Args:
            dump_file (Path, optional): Where to dump the csv. Defaults to Path("services.csv").
        """
        assert self.gatt_table is not None
        with open(dump_file, mode="w") as f:
            logger.debug(f"Dumping discovered BLE characteristics to {dump_file}")
            w = csv.writer(f, delimiter=",", quotechar='"', quoting=csv.QUOTE_MINIMAL)
            w.writerow(["handle", "description", "UUID", "properties", "value"])
            # For each service in table
            for s in self.gatt_table.services.values():
                w.writerow(["SERVICE", s.name, s.uuid.value, "SERVICE", "SERVICE"])
                # For each characteristic in service
                for c in s.chars.values():
                    w.writerow([c.handle, c.name, c.uuid.value, " ".join(c.props), c.value])
                    # For each descriptor in characteristic
                    for d in c.descriptors:
                        w.writerow([d.handle, "DESCRIPTOR", "", "", d.value])



# params.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue Sep  7 21:35:53 UTC 2021

# pylint: disable=missing-class-docstring

"""Updates to parameter definitions for GoPro BLE and WiFi commands for Open GoPro version 2.0"""

from open_gopro.constants import GoProEnum
from open_gopro.api.v1_0.params import ParamsV1_0


class ParamsV2_0(ParamsV1_0):
    class Resolution(GoProEnum):
        NOT_APPLICABLE = 0
        RES_4K = 1
        RES_2_7K = 4
        RES_2_7K_4_3 = 6
        RES_1080 = 9
        RES_4K_4_3 = 18
        RES_5_K_4_3 = 25
        RES_5_3_K = 100

    class Preset(GoProEnum):
        STANDARD = 0x00000000
        ACTIVITY = 0x00000001
        CINEMATIC = 0x00000002
        ULTRA_SLOMO = 0x00000004
        BASIC = 0x00000005
        PHOTO = 0x00010000
        LIVE_BURST = 0x00010001
        BURST_PHOTO = 0x00010002
        NIGHT_PHOTO = 0x00010003
        TIME_WARP = 0x00020000
        TIME_LAPSE = 0x00020001
        NIGHT_LAPSE = 0x00020002
        MAX_VIDEO = 0x00030000
        MAX_PHOTO = 0x00040000
        MAX_TIMEWARP = 0x00050000
        STANDARD_EB = 0x00080000
        ACTIVITY_EB = 0x00080001
        CINEMATIC_EB = 0x00080002
        SLOMO_EB = 0x00080003
        TRIPOD_4K = 0x00090000
        TRIPOD_5_3K = 0x00090001

    class LED(GoProEnum):
        ALL_ON = 3
        ALL_OFF = 4
        FRONT_OFF_ONLY = 5
        BLE_KEEP_ALIVE = 66

    class CameraControl(GoProEnum):
        IDLE = 0
        CAMERA = 1
        EXTERNAL = 2

    class VideoFOV(GoProEnum):
        WIDE = 0
        NARROW = 2
        SUPERVIEW = 3
        LINEAR = 4
        MAX_SUPERVIEW = 7
        LINEAR_HORIZON = 8

    class PhotoFOV(GoProEnum):
        NOT_APPLICABLE = 0
        WIDE = 101
        LINEAR = 102
        NARROW = 19
        MAX_SUPERVIEW = 100

    class MultishotFOV(GoProEnum):
        NOT_APPLICABLE = 0
        NARROW = 19
        MAX_SUPERVIEW = 100
        WIDE = 101

    class VideoPerformanceMode(GoProEnum):
        MAX_PERFORMANCE = 0
        EXTENDED_BATTERY = 1
        STATIONARY = 2



# wifi_commands.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue Sep  7 21:35:53 UTC 2021

"""Updates to WiFi API for Open GoPro version 2.0"""

from __future__ import annotations
import logging
from typing import Type

from open_gopro.communication_client import GoProWifi
from open_gopro.api.v1_0.wifi_commands import WifiCommandsV1_0, WifiSettingsV1_0
from open_gopro.api.v2_0.params import ParamsV2_0 as Params
from open_gopro.api.builders import WifiGetJsonNoParams, WifiGetJsonWithParams, WifiSetting
from open_gopro.constants import SettingId

logger = logging.getLogger(__name__)


class WifiCommandsV2_0(WifiCommandsV1_0):
    """All of the Wifi commands.

    To be used as a delegate for a GoProWifi to build commands

    All of these return a GoProResp

    Args:
        communicator (GoProWifi): [description]
    """

    def __init__(self, communicator: GoProWifi):
        super().__init__(communicator)

        self.set_third_party_client_info = WifiGetJsonNoParams(
            communicator, "/gopro/camera/analytics/set_client_info"
        )
        """Flag as third party app.

        Will return error if the camera is already encoding.

        Returns:
            GoProResp: command status
        """

        self.set_shutter_on = WifiGetJsonNoParams(communicator, "/gopro/camera/shutter/start")
        """Set the shutter on (i.e. start encoding).

        Returns:
            GoProResp: command status
        """

        self.set_shutter_off = WifiGetJsonNoParams(communicator, "/gopro/camera/shutter/stop")
        """Set the shutter off (i.e. stop encoding).

        Will return error tf this command is sent when the camera is not encoding.

        Returns:
            GoProResp: command status
        """

        self.set_camera_control = WifiGetJsonWithParams[Params.CameraControl](
            communicator, "/gopro/camera/control/set_ui_controller?p={}"
        )
        """Configure global behaviors by setting camera control (to i.e. Idle, External)

        Args:
            value (Params.CameraControl): desired camera control value

        Returns:
            GoProResp: command status
        """


class WifiSettingsV2_0(WifiSettingsV1_0):
    """Updates to WiFi settings for Open GoPro version 2.0"""

    def __init__(self, communicator: GoProWifi, params: Type[Params]):
        super().__init__(communicator, params, endpoint="gopro/camera/setting?setting={}&option={}")

        self.video_performance_mode = WifiSetting[Params.VideoPerformanceMode](
            communicator, SettingId.VIDEO_PERFORMANCE_MODE
        )
        """Video Performance Mode (extended battery, tripod, etc)."""



# ble_commands.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue Sep  7 21:35:52 UTC 2021

"""BLE API for Open GoPro version 1.0"""

from __future__ import annotations
import logging
from typing import Type

from construct import Int8ub, Int32ub, Int64ub, Flag, GreedyString, Struct, Padding, PaddedString, this, Hex

from open_gopro import proto
from open_gopro.responses import GoProResp
from open_gopro.communication_client import GoProBle
from open_gopro.api.builders import (
    BleStatus,
    BleSetting,
    BleWriteNoParamsCommand,
    BleWriteWithParamsCommand,
    RegisterUnregisterAll,
    BleReadCommand,
    BleProtoCommand,
    build_enum_adapter,
    DateTimeAdapter,
)
from open_gopro.constants import (
    ActionId,
    CmdId,
    QueryCmdId,
    SettingId,
    StatusId,
    UUID,
)
from .params import ParamsV1_0 as Params

logger = logging.getLogger(__name__)


class BleCommandsV1_0:
    # pylint: disable = missing-class-docstring, arguments-differ, useless-super-delegation, missing-return-doc
    """All of the BLE commands.

    To be used as a delegate for a GoProBle instance to build commands

    All of these return a GoProResp

    Args:
        communicator (GoProBle): GoPro BLE communicator that will send the commands
    """

    def __init__(self, communicator: GoProBle):

        self.set_shutter = BleWriteWithParamsCommand[Params.Shutter](
            communicator, UUID.CQ_COMMAND, CmdId.SET_SHUTTER, Int8ub
        )
        """Set shutter on or off.

        Args:
            value (Params.Shutter): on or off

        Returns:
            GoProResp: command status
        """

        self.power_down = BleWriteNoParamsCommand(communicator, UUID.CQ_COMMAND, CmdId.POWER_DOWN)
        """Power down the camera.

        Returns:
            GoProResp: command status
        """

        self.sleep = BleWriteNoParamsCommand(communicator, UUID.CQ_COMMAND, CmdId.SLEEP)
        """Put the camera in standby.

        Returns:
            GoProResp: command status
        """

        self.enable_wifi_ap = BleWriteWithParamsCommand[bool](
            communicator, UUID.CQ_COMMAND, CmdId.SET_WIFI, Int8ub
        )
        """Enable / disable the Wi-Fi Access Point.

        Args:
            value (bool): True to enable, False to disable

        Returns:
            GoProResp: command status
        """

        self.get_hardware_info = BleWriteNoParamsCommand(
            communicator,
            UUID.CQ_COMMAND,
            CmdId.GET_HW_INFO,
            response_parser=Struct(
                Padding(1),
                "model_number" / Int32ub,
                "model_name_len" / Int8ub,
                "model_name" / PaddedString(this.model_name_len, "utf-8"),
                Padding(1),
                "board_type" / Hex(Int32ub),
                "firmware_version_len" / Int8ub,
                "firmware_version" / PaddedString(this.firmware_version_len, "utf-8"),
                "serial_number_len" / Int8ub,
                "serial_number" / PaddedString(this.serial_number_len, "utf-8"),
                "ap_ssid_len" / Int8ub,
                "ap_ssid" / PaddedString(this.ap_ssid_len, "utf-8"),
                "ap_mac_len" / Int8ub,
                "ap_mac" / PaddedString(this.ap_mac_len, "utf-8"),
            ),
        )
        """Get hardware information.

        Returns:
            GoProResp: command status and hardware info (model name, etc.)
        """

        self.load_preset_group = BleWriteWithParamsCommand[Params.PresetGroup](
            communicator, UUID.CQ_COMMAND, CmdId.LOAD_PRESET_GROUP, Int32ub
        )
        """Load a Preset Group.

        Once complete, the most recently used preset in this group will be active.

        Args:
            value (Params.PresetGroup): preset group to load.

        Returns:
            GoProResp: command status
        """

        self.load_preset = BleWriteWithParamsCommand[Params.Preset](
            communicator, UUID.CQ_COMMAND, CmdId.LOAD_PRESET, Int32ub
        )
        """Load a Preset

        Args:
            value (Params.Preset): preset load.

        Returns:
            GoProResp: command status
        """

        self.set_third_party_client_info = BleWriteNoParamsCommand(
            communicator, UUID.CQ_COMMAND, CmdId.SET_THIRD_PARTY_CLIENT_INFO
        )
        """Flag as third party app.

        Returns:
            GoProResp: command status
        """

        self.get_open_gopro_api_version = BleWriteNoParamsCommand(
            communicator,
            UUID.CQ_COMMAND,
            CmdId.GET_THIRD_PARTY_API_VERSION,
            response_parser=Struct(Padding(1), "major" / Int8ub, Padding(1), "minor" / Int8ub),
        )
        """Get Open GoPro API version that is supported by the peer camera.

        Returns:
            GoProResp: command status and version (separate into major and minor fields)
        """

        class TurboMode(BleProtoCommand):
            def __call__(self, active: bool) -> GoProResp:
                return super().__call__(active)

        self.set_turbo_mode = TurboMode(
            communicator,
            UUID.CQ_COMMAND,
            CmdId.SET_TURBO_MODE,
            ActionId.SET_TURBO_MODE,
            proto.RequestSetTurboActive,
            proto.ResponseGeneric,
        )
        """Enable / disable turbo mode.

        Args:
            active (bool): True to enable, False to disable.

        Returns:
            TODO
        """

        # @proto_cmd(
        #     UUID.CQ_COMMAND,
        #     CmdId.GET_PRESET_STATUS,
        #     ActionId.GET_PRESET_STATUS,
        #     proto.RequestGetPresetStatus,
        #     proto.NotifyPresetStatus,
        # )
        # def get_preset_status(
        #     self,
        #     register_preset_status: List[Params.RegisterPresetStatus] = None,
        #     unregister_preset_status: List[Params.RegisterPresetStatus] = None,
        # ) -> GoProResp:
        #     """Get the preset status.

        #     Args:
        #         register_preset_status (List[Params.EnumRegisterPresetStatus], optional): [description]. Defaults to None.
        #         unregister_preset_status (List[Params.EnumRegisterPresetStatus], optional): [description]. Defaults to None.

        #     Returns:
        #         GoProResp: the response
        #     """

        self.get_wifi_ssid = BleReadCommand(
            communicator, UUID.WAP_SSID, Struct("ssid" / GreedyString("utf-8"))
        )
        """Get the Wifi SSID.

        Returns:
            GoProResp: command status and SSID
        """

        self.get_wifi_password = BleReadCommand(
            communicator, UUID.WAP_PASSWORD, Struct("password" / GreedyString("utf-8"))
        )
        """Get the Wifi password.

        Returns:
            GoProResp: command status and passworde
        """

        self.get_camera_statuses = BleWriteNoParamsCommand(
            communicator, UUID.CQ_QUERY, CmdId.GET_CAMERA_STATUSES
        )
        """Get all camera statuses.

        Returns:
            GoProResp: command status and current value of all statuses
        """

        self.get_camera_settings = BleWriteNoParamsCommand(
            communicator, UUID.CQ_QUERY, CmdId.GET_CAMERA_SETTINGS
        )
        """Get all camera settings.

        Returns:
            GoProResp: command status and current value of all settings
        """

        self.register_for_all_statuses = RegisterUnregisterAll(
            communicator,
            UUID.CQ_QUERY,
            CmdId.REGISTER_ALL_STATUSES,
            producer=(StatusId, QueryCmdId.STATUS_VAL_PUSH),
            action=RegisterUnregisterAll.Action.REGISTER,
        )
        """Register push notifications for all statuses

        Returns:
            GoProResp: command status and current value of all statuses
        """

        self.unregister_for_all_statuses = RegisterUnregisterAll(
            communicator,
            UUID.CQ_QUERY,
            CmdId.UNREGISTER_ALL_STATUSES,
            producer=(StatusId, QueryCmdId.STATUS_VAL_PUSH),
            action=RegisterUnregisterAll.Action.UNREGISTER,
        )
        """Unregister push notifications for all statuses

        Returns:
            GoProResp: command status
        """

        self.register_for_all_settings = RegisterUnregisterAll(
            communicator,
            UUID.CQ_QUERY,
            CmdId.REGISTER_ALL_SETTINGS,
            producer=(SettingId, QueryCmdId.STATUS_VAL_PUSH),
            action=RegisterUnregisterAll.Action.REGISTER,
        )
        """Register push notifications for all settings

        Returns:
            GoProResp: command status and current value of all settings
        """

        self.unregister_for_all_settings = RegisterUnregisterAll(
            communicator,
            UUID.CQ_QUERY,
            CmdId.UNREGISTER_ALL_SETTINGS,
            producer=(SettingId, QueryCmdId.STATUS_VAL_PUSH),
            action=RegisterUnregisterAll.Action.UNREGISTER,
        )
        """Unregister push notifications for all settings

        Returns:
            GoProResp: command status
        """

        # TODO
        # self.register_for_all_capabilities


class BleSettingsV1_0:
    # pylint: disable=missing-class-docstring, unused-argument
    """The collection of all Settings.

    To be used by a GoProBle delegate to build setting commands.

    Args:
        communicator (GoProBle): Adapter to read / write settings
        params: (Type[Params]): the set of parameters to use to build the settings
    """

    class Iterator:
        """Iterator to iterate through a BleSettings instance's attributes

        Does not include the 'communicator' instance.
        """

        def __init__(self, settings: "BleSettingsV1_0"):
            self._index = 0
            self._setting_attributes = list(settings.__dict__.values())[1:]  # Skip communicator

        def __next__(self) -> BleSetting:
            """Return next attribute

            Raises:
                StopIteration: Iteration has completed

            Returns:
                Setting: next attribute
            """
            if self._index < len(self._setting_attributes):
                setting = self._setting_attributes[self._index]
                self._index += 1
                return setting
            # End of Iteration
            raise StopIteration

    def __init__(self, communicator: GoProBle, params: Type[Params]):
        self.communicator = communicator

        self.resolution: BleSetting = BleSetting[Params.Resolution](
            self.communicator, SettingId.RESOLUTION, build_enum_adapter(params.Resolution)
        )
        """Resolution. Set with :py:class:`Params.Resolution`"""

        self.fps: BleSetting = BleSetting[Params.FPS](
            self.communicator, SettingId.FPS, build_enum_adapter(params.FPS)
        )
        """Frames per second. Set with :py:class:`Params.FPS`"""

        self.auto_off: BleSetting = BleSetting[Params.AutoOff](
            self.communicator, SettingId.AUTO_OFF, build_enum_adapter(params.AutoOff)
        )
        """Set the auto off time. Set with :py:class:`Params.AutoOff`"""

        self.video_field_of_view: BleSetting = BleSetting[Params.VideoFOV](
            self.communicator, SettingId.VIDEO_FOV, build_enum_adapter(params.VideoFOV)
        )
        """Video FOV. Set with :py:class:`Params.VideoFOV`"""

        self.photo_field_of_view: BleSetting = BleSetting[Params.PhotoFOV](
            self.communicator, SettingId.PHOTO_FOV, build_enum_adapter(params.PhotoFOV)
        )
        """Photo FOV. Set with :py:class:`Params.PhotoFOV`"""

        self.multi_shot_field_of_view: BleSetting = BleSetting[Params.MultishotFOV](
            self.communicator, SettingId.MULTI_SHOT_FOV, build_enum_adapter(params.MultishotFOV)
        )
        """Multi-shot FOV. Set with :py:class:`Params.Multishot`"""

        self.led: BleSetting = BleSetting[Params.LED](
            self.communicator, SettingId.LED, build_enum_adapter(params.LED)
        )
        """Set the LED options (or also send the BLE keep alive signal). Set with :py:class:`Params.LED`"""

        self.max_lens_mode: BleSetting = BleSetting[Params.MaxLensMode](
            self.communicator, SettingId.MAX_LENS_MOD, build_enum_adapter(params.MaxLensMode)
        )
        """Enable / disable max lens mod. Set with :py:class:`Params.MaxLensMode`"""

    def __iter__(self) -> Iterator:
        """Return an iterable of this instance's attributes

        Does not include the 'communicator' attribute

        Returns:
            Iterator: next attribute
        """
        return BleSettingsV1_0.Iterator(self)


class BleStatusesV1_0:
    """All of the BLE Statuses.

    To be used by a GoProBle delegate to build status commands.

    Args:
        communicator (GoProBle): Adapter to read / write settings
        params: (Type[Params]): the set of parameters to use to build the statuses
    """

    class Iterator:
        """Iterator to iterate through a BleStatuses instance's attributes

        Does not include the 'communicator' instance.
        """

        def __init__(self, statuses: "BleStatusesV1_0"):
            self._index = 0
            self._status_attributes = list(statuses.__dict__.values())[1:]  # Skip communicator

        def __next__(self) -> BleStatus:
            """Return next attribute

            Raises:
                StopIteration: Iteration has completed

            Returns:
                Setting: next attribute
            """
            if self._index < len(self._status_attributes):
                status = self._status_attributes[self._index]
                self._index += 1
                return status
            # End of Iteration
            raise StopIteration

    def __init__(self, communicator: GoProBle, params: Type[Params]) -> None:
        self.communicator = communicator

        self.batt_present: BleStatus = BleStatus(self.communicator, StatusId.BATT_PRESENT, Flag)
        """Is the system's internal battery present?"""

        self.batt_level: BleStatus = BleStatus(self.communicator, StatusId.BATT_LEVEL, Int8ub)
        """Rough approximation of internal battery level in bars."""

        self.ext_batt_present: BleStatus = BleStatus(self.communicator, StatusId.EXT_BATT_PRESENT, Flag)
        """Is an external battery connected?"""

        self.ext_batt_level: BleStatus = BleStatus(self.communicator, StatusId.EXT_BATT_LEVEL, Int8ub)
        """External battery power level in percent."""

        self.system_hot: BleStatus = BleStatus(self.communicator, StatusId.SYSTEM_HOT, Flag)
        """Is the system currently overheating?"""

        self.system_busy: BleStatus = BleStatus(self.communicator, StatusId.SYSTEM_BUSY, Flag)
        """Is the camera busy?"""

        self.quick_capture: BleStatus = BleStatus(self.communicator, StatusId.QUICK_CAPTURE, Flag)
        """Is quick capture feature enabled?"""

        self.encoding_active: BleStatus = BleStatus(self.communicator, StatusId.ENCODING, Flag)
        """Is the camera currently encoding (i.e. capturing photo / video)?"""

        self.lcd_lock_active: BleStatus = BleStatus(self.communicator, StatusId.LCD_LOCK_ACTIVE, Flag)
        """Is the LCD lock currently active?"""

        self.video_progress: BleStatus = BleStatus(self.communicator, StatusId.VIDEO_PROGRESS, Int32ub)
        """When encoding video, this is the duration (seconds) of the video so far; 0 otherwise."""

        self.wireless_enabled: BleStatus = BleStatus(self.communicator, StatusId.WIRELESS_ENABLED, Flag)
        """Are Wireless Connections enabled?"""

        self.pair_state: BleStatus = BleStatus(
            self.communicator, StatusId.PAIR_STATE, build_enum_adapter(params.PairState)
        )
        """What is the pair state?"""

        self.pair_type: BleStatus = BleStatus(
            self.communicator, StatusId.PAIR_TYPE, build_enum_adapter(params.PairType)
        )
        """The last type of pairing that the camera was engaged in."""

        self.pair_time: BleStatus = BleStatus(self.communicator, StatusId.PAIR_TIME, Int32ub)
        """	Time (milliseconds) since boot of last successful pairing complete action."""

        self.wap_scan_state: BleStatus = BleStatus(
            self.communicator, StatusId.WAP_SCAN_STATE, build_enum_adapter(params.WAPState)
        )
        """State of current scan for Wifi Access Points. Appears to only change for CAH-related scans."""

        self.wap_scan_time: BleStatus = BleStatus(self.communicator, StatusId.WAP_SCAN_TIME, Int8ub)
        """The time, in milliseconds since boot that the Wifi Access Point scan completed."""

        self.wap_prov_stat: BleStatus = BleStatus(
            self.communicator, StatusId.WAP_PROV_STAT, build_enum_adapter(params.WAPState)
        )
        """Wifi AP provisioning state."""

        self.remote_ctrl_ver: BleStatus = BleStatus(self.communicator, StatusId.REMOTE_CTRL_VER, Int8ub)
        """What is the remote control version?"""

        self.remote_ctrl_conn: BleStatus = BleStatus(self.communicator, StatusId.REMOTE_CTRL_CONN, Flag)
        """Is the remote control connected?"""

        self.pair_state2: BleStatus = BleStatus(self.communicator, StatusId.PAIR_STATE2, Int8ub)
        """Wireless Pairing State."""

        self.wlan_ssid: BleStatus = BleStatus(
            self.communicator, StatusId.WLAN_SSID, GreedyString(encoding="utf-8")
        )
        """Provisioned WIFI AP SSID. On BLE connection, value is big-endian byte-encoded int."""

        self.ap_ssid: BleStatus = BleStatus(
            self.communicator, StatusId.AP_SSID, GreedyString(encoding="utf-8")
        )
        """Camera's WIFI SSID. On BLE connection, value is big-endian byte-encoded int."""

        self.app_count: BleStatus = BleStatus(self.communicator, StatusId.APP_COUNT, Int8ub)
        """The number of wireless devices connected to the camera."""

        self.preview_enabled: BleStatus = BleStatus(self.communicator, StatusId.PREVIEW_ENABLED, Flag)
        """Is preview stream enabled?"""

        self.sd_status: BleStatus = BleStatus(
            self.communicator, StatusId.SD_STATUS, build_enum_adapter(params.SDStatus)
        )
        """Primary Storage Status."""

        self.photos_rem: BleStatus = BleStatus(self.communicator, StatusId.PHOTOS_REM, Int32ub)
        """How many photos can be taken before sdcard is full?"""

        self.video_rem: BleStatus = BleStatus(self.communicator, StatusId.VIDEO_REM, Int32ub)
        """How many minutes of video can be captured with current settings before sdcard is full?"""

        self.num_group_photo: BleStatus = BleStatus(self.communicator, StatusId.NUM_GROUP_PHOTO, Int32ub)
        """How many group photos can be taken with current settings before sdcard is full?"""

        self.num_group_video: BleStatus = BleStatus(self.communicator, StatusId.NUM_GROUP_VIDEO, Int32ub)
        """Total number of group videos on sdcard."""

        self.num_total_photo: BleStatus = BleStatus(self.communicator, StatusId.NUM_TOTAL_PHOTO, Int32ub)
        """Total number of photos on sdcard."""

        self.num_total_video: BleStatus = BleStatus(self.communicator, StatusId.NUM_TOTAL_VIDEO, Int32ub)
        """Total number of videos on sdcard."""

        self.date_time: BleStatus = BleStatus(
            self.communicator, StatusId.DATE_TIME, DateTimeAdapter(GreedyString(encoding="utf-8"))
        )
        """Current date/time (format: %YY%MM%DD%HH%MM%SS, all values in hex)."""

        self.ota_stat: BleStatus = BleStatus(
            self.communicator, StatusId.OTA_STAT, build_enum_adapter(params.OTAStatus)
        )
        """The current status of Over The Air (OTA) update."""

        self.download_cancel_pend: BleStatus = BleStatus(self.communicator, StatusId.DOWNLAD_CANCEL_PEND, Flag)
        """Is download firmware update cancel request pending?"""

        self.mode_group: BleStatus = BleStatus(self.communicator, StatusId.MODE_GROUP, Int8ub)
        """Current mode group (deprecated in HERO8)."""

        self.locate_active: BleStatus = BleStatus(self.communicator, StatusId.LOCATE_ACTIVE, Flag)
        """Is locate camera feature active?"""

        self.multi_count_down: BleStatus = BleStatus(self.communicator, StatusId.MULTI_COUNT_DOWN, Int32ub)
        """The current timelapse interval countdown value (e.g. 5...4...3...2...1...)."""

        self.space_rem: BleStatus = BleStatus(self.communicator, StatusId.SPACE_REM, Int64ub)
        """Remaining space on the sdcard in Kilobytes."""

        self.streaming_supp: BleStatus = BleStatus(self.communicator, StatusId.STREAMING_SUPP, Flag)
        """Is streaming supports in current recording/flatmode/secondary-stream?"""

        self.wifi_bars: BleStatus = BleStatus(self.communicator, StatusId.WIFI_BARS, Int8ub)
        """Wifi signal strength in bars."""

        self.current_time_ms: BleStatus = BleStatus(self.communicator, StatusId.CURRENT_TIME_MS, Int32ub)
        """System time in milliseconds since system was booted."""

        self.num_hilights: BleStatus = BleStatus(self.communicator, StatusId.NUM_HILIGHTS, Int8ub)
        """The number of hilights in encoding video (set to 0 when encoding stops)."""

        self.last_hilight: BleStatus = BleStatus(self.communicator, StatusId.LAST_HILIGHT, Int32ub)
        """Time since boot (msec) of most recent hilight in encoding video (set to 0 when encoding stops)."""

        self.next_poll: BleStatus = BleStatus(self.communicator, StatusId.NEXT_POLL, Int32ub)
        """The min time between camera status updates (msec). Do not poll for status more often than this."""

        self.analytics_rdy: BleStatus = BleStatus(
            self.communicator, StatusId.ANALYTICS_RDY, build_enum_adapter(params.AnalyticsState)
        )
        """The current state of camera analytics."""

        self.analytics_size: BleStatus = BleStatus(self.communicator, StatusId.ANALYTICS_SIZE, Int32ub)
        """The size (units??) of the analytics file."""

        self.in_context_menu: BleStatus = BleStatus(self.communicator, StatusId.IN_CONTEXT_MENU, Flag)
        """Is the camera currently in a contextual menu (e.g. Preferences)?"""

        self.timelapse_rem: BleStatus = BleStatus(self.communicator, StatusId.TIMELAPSE_REM, Int32ub)
        """How many min of Timelapse video can be captured with current settings before sdcard is full?"""

        self.exposure_type: BleStatus = BleStatus(
            self.communicator, StatusId.EXPOSURE_TYPE, build_enum_adapter(params.ExposureMode)
        )
        """Liveview Exposure Select Mode."""

        self.exposure_x: BleStatus = BleStatus(self.communicator, StatusId.EXPOSURE_X, Int8ub)
        """Liveview Exposure Select: y-coordinate (percent)."""

        self.exposure_y: BleStatus = BleStatus(self.communicator, StatusId.EXPOSURE_Y, Int8ub)
        """Liveview Exposure Select: y-coordinate (percent)."""

        self.gps_stat: BleStatus = BleStatus(self.communicator, StatusId.GPS_STAT, Flag)
        """Does the camera currently have a GPS lock?"""

        self.ap_state: BleStatus = BleStatus(self.communicator, StatusId.AP_STATE, Flag)
        """Is the Wifi radio enabled?"""

        self.int_batt_per: BleStatus = BleStatus(self.communicator, StatusId.INT_BATT_PER, Int8ub)
        """Internal battery level (percent)."""

        self.acc_mic_stat: BleStatus = BleStatus(
            self.communicator, StatusId.ACC_MIC_STAT, build_enum_adapter(params.ExposureMode)
        )
        """Microphone Accesstory status."""

        self.digital_zoom: BleStatus = BleStatus(self.communicator, StatusId.DIGITAL_ZOOM, Int8ub)
        """	Digital Zoom level (percent)."""

        self.wireless_band: BleStatus = BleStatus(
            self.communicator, StatusId.WIRELESS_BAND, build_enum_adapter(params.WifiBand)
        )
        """Wireless Band."""

        self.dig_zoom_active: BleStatus = BleStatus(self.communicator, StatusId.DIG_ZOOM_ACTIVE, Flag)
        """Is Digital Zoom feature available?"""

        self.mobile_video: BleStatus = BleStatus(self.communicator, StatusId.MOBILE_VIDEO, Flag)
        """Are current video settings mobile friendly? (related to video compression and frame rate)."""

        self.first_time: BleStatus = BleStatus(self.communicator, StatusId.FIRST_TIME, Flag)
        """Is the camera currently in First Time Use (FTU) UI flow?"""

        self.sec_sd_stat: BleStatus = BleStatus(
            self.communicator, StatusId.SEC_SD_STAT, build_enum_adapter(params.SDStatus)
        )
        """Secondary Storage Status (exclusive to Superbank)."""

        self.band_5ghz_avail: BleStatus = BleStatus(self.communicator, StatusId.BAND_5GHZ_AVAIL, Flag)
        """Is 5GHz wireless band available?"""

        self.system_ready: BleStatus = BleStatus(self.communicator, StatusId.SYSTEM_READY, Flag)
        """Is the system ready to accept commands?"""

        self.batt_ok_ota: BleStatus = BleStatus(self.communicator, StatusId.BATT_OK_OTA, Flag)
        """Is the internal battery charged sufficiently to start Over The Air (OTA) update?"""

        self.video_low_temp: BleStatus = BleStatus(self.communicator, StatusId.VIDEO_LOW_TEMP, Flag)
        """Is the camera getting too cold to continue recording?"""

        self.orientation: BleStatus = BleStatus(
            self.communicator, StatusId.ORIENTATION, build_enum_adapter(params.Orientation)
        )
        """The rotational orientation of the camera."""

        self.thermal_mit_mode: BleStatus = BleStatus(self.communicator, StatusId.THERMAL_MIT_MODE, Flag)
        """Can camera use high resolution/fps (based on temperature)?"""

        self.zoom_encoding: BleStatus = BleStatus(self.communicator, StatusId.ZOOM_ENCODING, Flag)
        """Is this camera capable of zooming while encoding (static value based on model, not settings)?"""

        self.flatmode_id: BleStatus = BleStatus(
            self.communicator, StatusId.FLATMODE_ID, build_enum_adapter(params.Flatmode)
        )
        """Current flatmode ID."""

        self.logs_ready: BleStatus = BleStatus(self.communicator, StatusId.LOGS_READY, Flag)
        """	Are system logs ready to be downloaded?"""

        self.timewarp_1x_active: BleStatus = BleStatus(self.communicator, StatusId.TIMEWARP_1X_ACTIVE, Flag)
        """Is Timewarp 1x active?"""

        self.video_presets: BleStatus = BleStatus(self.communicator, StatusId.VIDEO_PRESETS, Int32ub)
        """Current Video Preset (ID)."""

        self.photo_presets: BleStatus = BleStatus(self.communicator, StatusId.PHOTO_PRESETS, Int32ub)
        """Current Photo Preset (ID)."""

        self.timelapse_presets: BleStatus = BleStatus(self.communicator, StatusId.TIMELAPSE_PRESETS, Int32ub)
        """	Current Timelapse Preset (ID)."""

        self.presets_group: BleStatus = BleStatus(self.communicator, StatusId.PRESETS_GROUP, Int32ub)
        """Current Preset Group (ID)."""

        self.active_preset: BleStatus = BleStatus(self.communicator, StatusId.ACTIVE_PRESET, Int32ub)
        """Currently Preset (ID)."""

        self.preset_modified: BleStatus = BleStatus(self.communicator, StatusId.PRESET_MODIFIED, Int32ub)
        """Preset Modified Status, which contains an event ID and a preset (group) ID."""

        self.live_burst_rem: BleStatus = BleStatus(self.communicator, StatusId.LIVE_BURST_REM, Int32ub)
        """How many Live Bursts can be captured before sdcard is full?"""

        self.live_burst_total: BleStatus = BleStatus(self.communicator, StatusId.LIVE_BURST_TOTAL, Int32ub)
        """Total number of Live Bursts on sdcard."""

        self.capt_delay_active: BleStatus = BleStatus(self.communicator, StatusId.CAPT_DELAY_ACTIVE, Flag)
        """Is Capture Delay currently active (i.e. counting down)?"""

        self.media_mod_mic_stat: BleStatus = BleStatus(
            self.communicator,
            StatusId.MEDIA_MOD_MIC_STAT,
            build_enum_adapter(params.MediaModMicStatus),
        )
        """Media mod State."""

        self.timewarp_speed_ramp: BleStatus = BleStatus(
            self.communicator, StatusId.TIMEWARP_SPEED_RAMP, build_enum_adapter(params.TimeWarpSpeed)
        )
        """Time Warp Speed."""

        self.linux_core_active: BleStatus = BleStatus(self.communicator, StatusId.LINUX_CORE_ACTIVE, Flag)
        """Is the system's Linux core active?"""

        self.camera_lens_type: BleStatus = BleStatus(
            self.communicator, StatusId.CAMERA_LENS_TYPE, build_enum_adapter(params.MaxLensMode)
        )
        """Camera lens type (reflects changes to setting 162)."""

        self.video_hindsight: BleStatus = BleStatus(self.communicator, StatusId.VIDEO_HINDSIGHT, Flag)
        """Is Video Hindsight Capture Active?"""

        self.scheduled_preset: BleStatus = BleStatus(self.communicator, StatusId.SCHEDULED_PRESET, Int32ub)
        """Scheduled Capture Preset ID."""

        self.scheduled_capture: BleStatus = BleStatus(self.communicator, StatusId.SCHEDULED_CAPTURE, Flag)
        """Is Scheduled Capture set?"""

        self.creating_preset: BleStatus = BleStatus(self.communicator, StatusId.CREATING_PRESET, Flag)
        """Is the camera in the process of creating a custom preset?"""

        self.media_mod_stat: BleStatus = BleStatus(
            self.communicator, StatusId.MEDIA_MOD_STAT, build_enum_adapter(params.MediaModStatus)
        )
        """Media Mode Status (bitmasked)."""

        self.turbo_mode: BleStatus = BleStatus(self.communicator, StatusId.TURBO_MODE, Flag)
        """Is Turbo Transfer active?"""

    def __iter__(self) -> Iterator:
        """Return an iterable of this instance's attributes

        Does not include the 'communicator' attribute

        Returns:
            Iterator: next attribute
        """
        return BleStatusesV1_0.Iterator(self)



# photo.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Wed, Sep  1, 2021  5:05:45 PM

"""Entrypoint for taking a picture demo."""

import sys
import logging
import argparse
from pathlib import Path
from typing import Tuple, Optional

from rich.console import Console

from open_gopro import GoPro
from open_gopro.util import setup_logging

logger = logging.getLogger(__name__)
console = Console()  # rich consoler printer


def main(identifier: Optional[str], log_location: Path, output_location: Path) -> int:
    """Main program functionality

    Args:
        identifier (Optional[str]): device to connect to
        log_location (Path): file to write detailed log
        output_location (Path): name of photo file

    Returns:
        int: program return code
    """
    global logger
    logger = setup_logging(logger, log_location)

    gopro: Optional[GoPro] = None
    return_code = 0
    try:
        with GoPro(identifier) as gopro:
            # Turn off the shutter if we are currently encoding
            if gopro.is_encoding:
                assert gopro.ble_command.set_shutter(gopro.params.Shutter.OFF).is_ok

            # Ensure we are in a mode that can enter photo preset
            if gopro.version >= 2.0:
                assert gopro.ble_setting.video_performance_mode.set(
                    gopro.params.VideoPerformanceMode.MAX_PERFORMANCE
                ).is_ok
                assert gopro.ble_setting.max_lens_mode.set(gopro.params.MaxLensMode.DEFAULT).is_ok
            assert gopro.ble_command.set_turbo_mode(False).is_ok
            assert gopro.ble_command.load_preset(gopro.params.Preset.PHOTO)

            console.print("Capturing a photo...")
            # Get the media list before
            media_set_before = set(x["n"] for x in gopro.wifi_command.get_media_list()["media"][0]["fs"])
            # Take a photo
            assert gopro.ble_command.load_preset(gopro.params.Preset.PHOTO).is_ok
            assert gopro.ble_command.set_shutter(gopro.params.Shutter.ON).is_ok

            console.print("Downloading the photo...")
            # Get the media list after
            media_set_after = set(x["n"] for x in gopro.wifi_command.get_media_list()["media"][0]["fs"])
            # The photo (is most likely) the difference between the two dicts as sets
            photo = media_set_after.difference(media_set_before).pop()
            # Download the photo
            gopro.wifi_command.download_file(camera_file=photo, local_file=output_location)
            console.print(
                f"Success!! :smiley: File has been downloaded to {output_location}", style="bold green"
            )

    except Exception as e:  # pylint: disable=broad-except
        logger.error(repr(e))
        return_code = 1
    except KeyboardInterrupt:
        logger.warning("Received keyboard interrupt. Shutting down...")
    finally:
        if gopro is not None:
            gopro.close()
        console.print("Exiting...")
        return return_code  # pylint: disable=lost-exception


def parse_arguments() -> Tuple[str, Path, Path]:
    """Parse command line arguments

    Returns:
        Tuple[str, Path, Path]: (identifier, path to save log, path to store photo)
    """
    parser = argparse.ArgumentParser(description="Connect to a GoPro camera, take a photo, then download it.")
    parser.add_argument(
        "-i",
        "--identifier",
        type=str,
        help="Last 4 digits of GoPro serial number, which is the last 4 digits of the default camera SSID. If \
            not used, first discovered GoPro will be connected to",
        default=None,
    )
    parser.add_argument(
        "-l",
        "--log",
        type=Path,
        help="Location to store detailed log",
        default=Path("photo.log"),
    )
    parser.add_argument(
        "-o",
        "--output",
        type=Path,
        help="Where to write the photo to. If not set, write to 'photo.jpg'",
        default=Path("photo.jpg"),
    )
    args = parser.parse_args()

    return args.identifier, args.log, args.output


def entrypoint() -> None:
    """Entrypoint for setup.py"""
    sys.exit(main(*parse_arguments()))


if __name__ == "__main__":
    entrypoint()



# __init__.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue Sep  7 21:35:53 UTC 2021

"""Top level API module. Used to abstract all API versions."""

from typing import Dict, Type, Union

from open_gopro.api.v1_0.ble_commands import BleCommandsV1_0, BleSettingsV1_0, BleStatusesV1_0
from open_gopro.api.v1_0.wifi_commands import WifiSettingsV1_0, WifiCommandsV1_0
from open_gopro.api.v1_0.params import ParamsV1_0
from open_gopro.api.v2_0.ble_commands import BleCommandsV2_0, BleSettingsV2_0, BleStatusesV2_0
from open_gopro.api.v2_0.wifi_commands import WifiSettingsV2_0, WifiCommandsV2_0
from open_gopro.api.v2_0.params import ParamsV2_0
from .v1_0.api import ApiV1_0
from .v2_0.api import ApiV2_0

# Version-agnostic type definitions
Api = Union[ApiV1_0, ApiV2_0]
BleCommands = Union[BleCommandsV1_0, BleCommandsV2_0]
BleSettings = Union[BleSettingsV1_0, BleSettingsV2_0]
BleStatuses = Union[BleStatusesV1_0, BleStatusesV2_0]
WifiSettings = Union[WifiSettingsV1_0, WifiSettingsV2_0]
WifiCommands = Union[WifiCommandsV1_0, WifiCommandsV2_0]
Params = Union[ParamsV1_0, ParamsV2_0]

api_versions: Dict[str, Type[Api]] = {
    api.version: api
    for api in [
        ApiV1_0,
        ApiV2_0,
    ]
}



# controller.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue Sep  7 21:35:53 UTC 2021

"""Wifi Controller Interface Definition."""

import logging
from abc import ABC, abstractmethod
from enum import IntEnum, auto
from typing import Optional, List, Tuple

logger = logging.getLogger(__name__)


class SsidState(IntEnum):
    """Current state of the SSID"""

    ESTABLISHING = auto()
    CONNECTED = auto()
    DISCONNECTED = auto()


class WifiController(ABC):
    """Interface definition for a Wifi driver to be used by GoPro."""

    @abstractmethod
    def connect(self, ssid: str, password: str, timeout: float = 15) -> bool:
        """Connect to a network.

        Args:
            ssid (str): SSID of network to connect to
            password (str): password of network to connect to
            timeout (float, optional): Time before considering connection failed (in seconds). Defaults to 15.

        Returns:
            bool: True if successful, False otherwise
        """
        raise NotImplementedError

    @abstractmethod
    def disconnect(self) -> bool:
        """Disconnect from a network.

        Returns:
            bool: True if successful, False otherwise
        """
        raise NotImplementedError

    @abstractmethod
    def current(self) -> Tuple[Optional[str], SsidState]:
        """Return the SSID and state of the current network.

        Returns:
            Tuple[Optional[str], SsidState]: Tuple of SSID str and state. If SSID is None,
            there is no current connection.
        """
        raise NotImplementedError

    @abstractmethod
    def interfaces(self) -> List[str]:
        """Return a list of wireless adapters.

        Returns:
            List[str]: adapters
        """
        raise NotImplementedError

    @abstractmethod
    def interface(self, interface: Optional[str]) -> Optional[str]:
        """Get or set the currently used wireless adapter.

        Args:
            interface (str, optional): Get if the interface parameter is None. Set otherwise. Defaults to None.

        Returns:
            Optional[str]: Name of interface if get. None if set.
        """
        raise NotImplementedError

    @abstractmethod
    def power(self, power: bool) -> bool:
        """Enable / disable the wireless driver.

        Args:
            power (bool, optional): Enable if True. Disable if False.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def is_on(self) -> bool:
        """Is the wireless driver currently enabled.

        Returns:
            bool: True if yes. False if no.
        """
        raise NotImplementedError



# controller.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue Sep  7 21:35:53 UTC 2021

"""BLE Controller Interface Definition."""

import logging
from abc import ABC, abstractmethod
from typing import Callable, Generic, Pattern, TypeVar

from .services import AttributeTable

logger = logging.getLogger(__name__)

BleDevice = TypeVar("BleDevice")
BleHandle = TypeVar("BleHandle")
NotiHandlerType = Callable[[int, bytearray], None]
DisconnectHandlerType = Callable[[BleDevice], None]


class BLEController(ABC, Generic[BleDevice, BleHandle]):
    """Interface definition for a BLE driver to be used by GoPro."""

    @abstractmethod
    def read(self, handle: BleHandle, uuid: str) -> bytearray:
        """Read a bytestream response from a UUID.

        Args:
            handle (BleHandle): handle to pair to
            uuid (UUID): UUID to read from

        Returns:
            bytearray: response
        """
        raise NotImplementedError

    @abstractmethod
    def write(self, handle: BleHandle, uuid: str, data: bytearray) -> None:
        """Write a bytestream to a UUID.

        Args:
            handle (BleHandle): handle to pair to
            uuid (UUID): UUID to write to
            data (bytearray): bytestream to write

        Returns:
            bytearray: response
        """
        raise NotImplementedError

    @abstractmethod
    def scan(self, token: Pattern, timeout: int = 5) -> BleDevice:
        """Scan BLE device with a regex in it's device name.

        Args:
            token (Pattern): Regex to scan for
            timeout (int, optional): [description]. Defaults to 5.

        Returns:
            Device: discovered device (shall not be multiple devices)
        """
        raise NotImplementedError

    @abstractmethod
    def connect(self, disconnect_cb: DisconnectHandlerType, device: BleDevice, timeout: int = 15) -> BleHandle:
        """Connect to a BLE device.

        Args:
            disconnect_cb (DisconnectHandlerType): function to call when disconnect is received
            device (BleDevice): device to connect to
            timeout (int, optional): How long to attempt connecting before giving up. Defaults to 15.

        Returns:
            handle: handle that has been connected to

        Returns:
            BleHandle: Handle to identify newly connected device
        """
        raise NotImplementedError

    @abstractmethod
    def pair(self, handle: BleHandle) -> None:
        """Pair to an already connected handle.

        Args:
            handle (BleHandle): handle to pair to
        """
        raise NotImplementedError

    @abstractmethod
    def enable_notifications(self, handle: BleHandle, handler: NotiHandlerType) -> None:
        """Enable notifications for all notifiable characteristics.

        The handler is used to register for notifications. It will be called when a a notification
        is received.

        Args:
            handle (BleHandle): handle to enable notifications on
            handler (Callable): notification handler
        """
        raise NotImplementedError

    @abstractmethod
    def discover_chars(self, handle: BleHandle) -> AttributeTable:
        """Discover all characteristics for a connected handle.

        Args:
            handle (BleHandle): handle to discover on

        Returns:
            BleHandle: handle with updated gatt_table attribute
        """
        raise NotImplementedError

    @abstractmethod
    def disconnect(self, handle: BleHandle) -> None:
        """Terminate the BLE connection.

        Args:
            handle (BleHandle): handle to disconnect from
        """
        raise NotImplementedError



# wifi_commands.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue Sep  7 21:35:52 UTC 2021

"""WiFi API for Open GoPro version 1.0"""

from __future__ import annotations
import logging
from pathlib import Path
from typing import Any, Optional, Type, Dict

from open_gopro.communication_client import GoProWifi
from open_gopro.constants import SettingId, StatusId
from open_gopro.responses import BytesParserBuilder, JsonParser
from open_gopro.api.builders import WifiSetting, WifiGetJsonNoParams, WifiGetJsonWithParams, WifiGetBinary
from .params import ParamsV1_0 as Params

logger = logging.getLogger(__name__)


class WifiCommandsV1_0:
    # pylint: disable = missing-class-docstring, arguments-differ, useless-super-delegation, missing-return-doc
    """All of the Wifi commands.

    To be used as a delegate for a GoProWifi to build commands

    All of these return a GoProResp

    Args:
        communicator (GoProWifi):  Adapter to read / write commands
    """

    # TODO refactor this to reuse response parsing
    class _ParseCameraState(JsonParser):
        """Additional parsing to do on received camera state."""

        def parse(
            self, buf: Dict[str, Any], additional_parsers: Dict[Any, BytesParserBuilder] = None
        ) -> Dict[Any, Any]:
            """Parse the raw state values into user friendly types / values.

            Args:
                buf (Dict[str, Any]): input dict to parse
                additional_parsers (Dict[Any, BytesParserBuilder], optional): additional parsers to be used. Defaults to None

            Returns:
                Dict[Any, Any]: parsed output dict
            """
            assert additional_parsers is not None
            parsed: Dict[Any, Any] = {}
            # Parse status and settings values into nice human readable things
            for (name, id_map) in [("status", StatusId), ("settings", SettingId)]:
                for k, v in buf[name].items():
                    identifier = id_map(int(k))
                    try:
                        parser = additional_parsers[identifier]
                        val = parser.parse(parser.build(v))
                    except KeyError:
                        logger.warning(f"unparsed {name}: {identifier}")
                        val = v
                    except ValueError:
                        logger.warning(f"{identifier} does not contain a value {v}")
                        val = v

                    parsed[identifier] = val

            return parsed

    def __init__(self, communicator: GoProWifi):
        self.communicator = communicator

        class CameraFileToLocalFile(WifiGetBinary):
            def __call__(self, /, camera_file: str, local_file: Optional[Path] = None) -> Path:
                return super().__call__(camera_file=camera_file, local_file=local_file or camera_file)

        # ======================================== Commands

        self.set_digital_zoom = WifiGetJsonWithParams[int](
            communicator, "gopro/camera/digital_zoom?percent={}"
        )
        """Set digital zoom in percent.

        Args:
            value (int): Desired zoom as a percentage

        Returns:
            GoProResp: command status
        """

        self.get_camera_state = WifiGetJsonNoParams(
            communicator, "gopro/camera/state", response_parser=WifiCommandsV1_0._ParseCameraState()
        )
        """Get camera status and settings.

        Returns:
            GoProResp: command status
        """

        self.set_keep_alive = WifiGetJsonNoParams(communicator, "/gopro/camera/keep_alive")
        """Send the keep alive signal to maintain the connection.

        Returns:
            GoProResp: command status
        """

        self.get_gpmf_data = CameraFileToLocalFile(communicator, "gopro/media/gpmf?path=100GOPRO/{}")
        """Get GPMF data for a file.

        If local_file is none, the output location will be the same name as the camera_file.

        Args:
            camera_file (str): filename on camera to operate on
            local_file (Optional[Path], optional): Location on computer to write output. Defaults to None.

        Returns:
            Path: Path to local_file that output was written to
        """

        self.get_media_info = WifiGetJsonWithParams[str](communicator, "gopro/media/info?path=100GOPRO/{}")
        """Get media info for a file.

        Args:
            value (str): Media file to get info for

        Returns:
            GoProResp: command status and media info as JSON
        """

        self.get_media_list = WifiGetJsonNoParams(communicator, "gopro/media/list")
        """Get a list of media on the camera.

        Returns:
            GoProResp: command status and media list as JSON
        """

        self.get_screennail = CameraFileToLocalFile(communicator, "gopro/media/screennail?path=100GOPRO/{}")
        """Get screennail for a file.

        If local_file is none, the output location will be the same name as the camera_file.

        Args:
            camera_file (str): filename on camera to operate on
            local_file (Optional[Path], optional): Location on computer to write output. Defaults to None.

        Returns:
            Path: Path to local_file that output was written to
        """

        self.get_thumbnail = CameraFileToLocalFile(communicator, "gopro/media/thumbnail?path=100GOPRO/{}")
        """Get thumbnail for a file.

        If local_file is none, the output location will be the same name as the camera_file.

        Args:
            camera_file (str): filename on camera to operate on
            local_file (Optional[Path], optional): Location on computer to write output. Defaults to None.

        Returns:
            Path: Path to local_file that output was written to
        """

        self.set_turbo_mode = WifiGetJsonWithParams[Params.Toggle](
            communicator, "/gopro/media/turbo_transfer?p={}"
        )
        """Enable or disable Turbo transfer mode.

        Args:
            value (Params.Toggle): enable / disable turbo mode

        Returns:
            GoProResp: command status
        """

        self.get_open_gopro_api_version = WifiGetJsonNoParams(communicator, "gopro/version")
        """Get Open GoPro version.

        Returns:
            GoProResp: command status and version separated into major and minor fields
        """

        self.get_preset_status = WifiGetJsonNoParams(communicator, "gopro/camera/presets/get")
        """Get current Preset status.

        Returns:
            GoProResp: command status and current preset status as JSON
        """

        self.set_preset = WifiGetJsonWithParams[Params.Preset](communicator, "gopro/camera/presets/load?id={}")
        """Set the active preset."""

        self.set_preset_group = WifiGetJsonWithParams[Params.PresetGroup](
            communicator, "gopro/camera/presets/set_group?id={}"
        )
        """Set the active preset group.

        The most recently used Preset in this group will be set.

        Args:
            value (Params.PresetGroup): desired Preset Group

        Returns:
            GoProResp: command status
        """

        self.start_preview_stream = WifiGetJsonNoParams(communicator, "gopro/camera/stream/start")
        """Start the preview stream.

        Returns:
            GoProResp: command status
        """

        self.stop_preview_stream = WifiGetJsonNoParams(communicator, "gopro/camera/stream/stop")
        """Stop the preview stream.

        Returns:
            GoProResp: command status
        """

        self.get_telemetry = CameraFileToLocalFile(communicator, "gopro/media/telemetry?path=100GOPRO/{}")
        """Download the telemetry data for a camera file and store in a local file.

        If local_file is none, the output location will be the same name as the camera_file.

        Args:
            camera_file (str): filename on camera to operate on
            local_file (Optional[Path], optional): Location on computer to write output. Defaults to None.

        Returns:
            Path: Path to local_file that output was written to
        """

        self.download_file = CameraFileToLocalFile(communicator, "videos/DCIM/100GOPRO/{}")
        """Download a video from the camera to a local file.

        If local_file is none, the output location will be the same name as the camera_file.

        Args:
            camera_file (str): filename on camera to operate on
            local_file (Optional[Path], optional): Location on computer to write output. Defaults to None.

        Returns:
            Path: Path to local_file that output was written to
        """


# =====================================================Settings============================================


class WifiSettingsV1_0:
    # pylint: disable=missing-class-docstring
    # pylint: disable=unused-argument
    """The collection of all Settings.

    Args:
        communicator (GoProWifi): Adapter to read / write settings
        params: (Type[Params]): the set of parameters to use to build the settings
        endpoint (str): the endpoint to write to to set a setting
    """

    class Iterator:
        """Iterator to iterate through a BleSettings instance's attributes

        Does not include the 'communicator' instance.
        """

        def __init__(self, settings: "WifiSettingsV1_0"):
            self._index = 0
            self._setting_attributes = list(settings.__dict__.values())[1:]  # Skip communicator

        def __next__(self) -> WifiSetting:
            """Return the next attribute

            Raises:
                StopIteration: Nothing left to iterate

            Returns:
                Setting: Instance of setting
            """
            if self._index < len(self._setting_attributes):
                setting = self._setting_attributes[self._index]
                self._index += 1
                return setting
            # End of Iteration
            raise StopIteration

    def __init__(
        self,
        communicator: GoProWifi,
        params: Type[Params],
        endpoint: str = "/gopro/camera/setting?setting_id={}&opt_value={}",
    ):
        self.endpoint = endpoint
        self.params = params  # Not currently used but will be in the future

        self.resolution = WifiSetting[Params.Resolution](communicator, SettingId.RESOLUTION)
        """Resolution. Set with :py:class:`open_gopro.Params.Resolution`"""

        self.fps = WifiSetting[Params.FPS](communicator, SettingId.FPS)
        """Frames per second. Set with :py:class:`open_gopro.Params.FPS`"""

        self.auto_off = WifiSetting[Params.AutoOff](communicator, SettingId.AUTO_OFF)
        """Set the auto off time. Set with :py:class:`Params.AutoOff`"""

        self.video_field_of_view = WifiSetting[Params.VideoFOV](communicator, SettingId.VIDEO_FOV)
        """Video FOV. Set with :py:class:`open_gopro.Params.FieldOfView`"""

        self.photo_field_of_view = WifiSetting[Params.PhotoFOV](communicator, SettingId.PHOTO_FOV)
        """Photo FOV. Set with :py:class:`open_gopro.Params.FieldOfView`"""

        self.multi_shot_field_of_view = WifiSetting[Params.MultishotFOV](
            communicator, SettingId.MULTI_SHOT_FOV
        )
        """Multi-shot FOV. Set with :py:class:`open_gopro.Params.FieldOfView`"""

        self.max_lens_mode = WifiSetting[Params.MaxLensMode](communicator, SettingId.MAX_LENS_MOD)
        """Enable / disable max lens mod. Set with :py:class:`open_gopro.Params.MaxLensMode`"""

    def __iter__(self) -> Iterator:
        """Return an iterable of this instance's attributes

        Does not include the 'communicator' attribute

        Returns:
            Iterator: next attribute
        """
        return WifiSettingsV1_0.Iterator(self)



# responses.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Wed, Sep  1, 2021  5:05:49 PM

"""Any responses that are returned from GoPro commands."""

import enum
import json
import logging
from collections import defaultdict
from typing import (
    Any,
    Dict,
    Callable,
    List,
    Optional,
    Union,
    Iterator,
    Type,
    ItemsView,
    ValuesView,
    KeysView,
)
from typing_extensions import Protocol

import requests
from construct import Construct

from open_gopro.exceptions import ResponseParseError
from open_gopro.constants import (
    ActionId,
    StatusId,
    CmdId,
    ErrorCode,
    SettingId,
    QueryCmdId,
    ResponseType,
    CmdType,
)
from open_gopro.util import scrub
from open_gopro.ble import UUID

logger = logging.getLogger(__name__)

BytesParser = Construct
BytesBuilder = Construct
BytesParserBuilder = Construct
StringBuilder = Callable[[Any], str]


class JsonParser(Protocol):
    """Protocol definition for a non-Construct parser to be used by GoProResp for parsing JSON input into  JSON output."""

    def parse(
        self, buf: Dict[str, Any], additional_parsers: Dict[Any, BytesParserBuilder] = None
    ) -> Dict[Any, Any]:
        """Parse JSON input into JSON dict

        Args:
            buf (Dict[str, Any]): JSON input
            additional_parsers (Dict[Any, BytesParserBuilder], optional): additional parsers. Defaults to None.

        Returns:
            Dict[Any, Any]: JSON response
        """
        ...


Parser = Union[BytesParser, JsonParser]

# Type of data that will be parsed into a GoPro Response.
InputType = Optional[Union[bytearray, Dict[str, Any]]]


ParserMapType = Dict[ResponseType, Parser]

CONT_MASK = 0b10000000
HDR_MASK = 0b01100000
GEN_LEN_MASK = 0b00011111
EXT_13_BYTE0_MASK = 0b00011111

id_map: Dict[UUID, Callable] = {
    UUID.CQ_SETTINGS_RESP: SettingId,
    UUID.CQ_COMMAND_RESP: CmdId,
    UUID.CQ_QUERY_RESP: QueryCmdId,
}

response_map: Dict[UUID, UUID] = {
    UUID.CQ_SETTINGS: UUID.CQ_SETTINGS_RESP,
    UUID.CQ_COMMAND: UUID.CQ_COMMAND_RESP,
    UUID.CQ_QUERY: UUID.CQ_QUERY_RESP,
}


class GoProResp:
    """A flexible object to be used to encapsulate all GoPro responses.

    It can be instantiated with varying levels of information and filled out as more is received.
    The end user should never need to create a response and will only be consuming them.

    It is mostly a wrapper around a JSON-like dictionary (GoProResp.data)
    For example, first send the command and store the response:

    >>> response = ble_setting.resolution.get_value()

    Now let's inspect the responses various attributes / properties:

    >>> print(response.status)
    ErrorCode.SUCCESS
    >>> print(response.is_ok)
    True
    >>> print(response.id)
    QueryCmdId.GET_SETTING_VAL
    >>> print(response.cmd)
    QueryCmdId.GET_SETTING_VAL
    >>> print(response.uuid)
    UUID.CQ_QUERY_RESP

    Now let's print it as (as JSON):

    >>> print(response.data)
    {
        "status": "SUCCESS",
        "id": "UUID.CQ_QUERY_RESP::QueryCmdId.GET_SETTING_VAL",
        "SettingId.RESOLUTION": [
            "RES_1080"
        ]
    }


    Args:
        info (List[ResponseType]): A list of all information known about the response.
        status (ErrorCode, optional): A status if known at time of instantiation. Defaults to ErrorCode.SUCCESS.
        raw_packet (InputType, optional): The unparsed input if known at time of instantiation. Defaults to None.
    """

    class _State(enum.Enum):
        """Describes the state of the GoProResp."""

        INITIALIZED = enum.auto()
        ACCUMULATED = enum.auto()
        PARSED = enum.auto()
        ERROR = enum.auto()

    def __init__(
        self,
        parsers: ParserMapType,
        info: List[ResponseType],
        status: ErrorCode = ErrorCode.SUCCESS,
        raw_packet: InputType = None,
    ) -> None:
        # A list describing all of the currently known information about the response.
        # This will be appended to as more information is discovered. The various properties of GoProResp will
        # use this list to parse out their relevant information.
        self._info: List[ResponseType] = info
        # Parsers to use to parse this response
        self._parsers = parsers

        self.status: ErrorCode = status
        """Status of the response"""

        # Start with empty list as default value in case we need to append.
        # If we end up not needing a list, we will just overwrite the default
        self.data: Dict[Any, Any] = defaultdict(list)
        """Response data which is really JSON data stored as a dict"""

        self._raw_packet = raw_packet
        self._bytes_remaining = 0
        self._state: GoProResp._State = GoProResp._State.INITIALIZED

    @classmethod
    def _from_write_command(cls, parsers: ParserMapType, uuid: UUID, data: bytes) -> "GoProResp":
        """Build a GoProResp from a write command.

        This will discover the expected response UUID from the command UUID as well as parse the ID
        from the bytestream.

        Args:
            uuid (UUID): UUID that write command is writing to
            data (bytes): bytestream of the command

        Returns:
            GoProResp: created instance
        """
        # Find expected response and ID from the data
        response_uuid = response_map[uuid]
        cmd_id = id_map[response_uuid](data[1])
        info = [response_uuid, cmd_id]
        # If this is a protobuf command, it also has an action id
        if cmd_id.value >= 0xF0:
            info.append(ActionId(data[2]))
        return cls(parsers, info=info)

    @classmethod
    def _from_read_response(cls, parsers: ParserMapType, uuid: UUID, data: bytearray) -> "GoProResp":
        """Build a GoProResp from a read response.

        Args:
            uuid (UUID): UUID that read command was received from
            data (bytes): received bytestream

        Returns:
            GoProResp: created instance
        """
        resp = cls(parsers, info=[uuid], status=ErrorCode.SUCCESS, raw_packet=data)
        resp._parse()
        return resp

    @classmethod
    def _from_http_response(cls, parsers: ParserMapType, response: requests.models.Response) -> "GoProResp":
        """Build a GoProResp from an HTTP response from the requests package.

        Args:
            response (requests.models.Response): HTTP response

        Returns:
            GoProResp: created instance
        """
        resp = cls(
            parsers,
            info=[response.request.path_url.strip("/")],
            status=ErrorCode.SUCCESS if response.ok else ErrorCode.ERROR,
            raw_packet=response.json(),
        )
        resp._parse()
        return resp

    def __getitem__(self, key: Any) -> Any:  # pylint: disable=missing-return-doc
        return self.data[key]

    def __contains__(self, item: Any) -> bool:  # pylint: disable=missing-return-doc
        return item in self.data

    def __iter__(self) -> Iterator:  # pylint: disable=missing-return-doc
        return iter(self.data)

    def __str__(self) -> str:  # pylint: disable=missing-return-doc
        return json.dumps(
            {
                "status": self.status.name,
                "id": "::".join([str(x) for x in self._info]),
                **{str(k): v for k, v in self.data.items()},
            },
            default=str,
            indent=4,
        )

    def items(self) -> ItemsView[Any, Any]:
        """Pass-through to access data dict's "items" method

        Returns:
            ItemsView[Any, Any]: [description]
        """
        return self.data.items()

    def keys(self) -> KeysView[Any]:
        """Pass-through to access data dict's "keys" method

        Returns:
            ItemsView[Any, Any]: [description]
        """
        return self.data.keys()

    def values(self) -> ValuesView[Any]:
        """Pass-through to access data dict's "values" method

        Returns:
            ItemsView[Any, Any]: [description]
        """
        return self.data.values()

    @property
    def is_received(self) -> bool:
        """Has the response been completely received?

        Returns:
            bool: True if completely received, False if not
        """
        return self._state is not GoProResp._State.INITIALIZED

    @property
    def is_parsed(self) -> bool:
        """Has the response been successfully parsed?

        Returns:
            bool: True if it has been parsed, False if not
        """
        return self._state is GoProResp._State.PARSED

    @property
    def is_ok(self) -> bool:
        """Are there any errors in this response?

        Returns:
            bool: True if the response is ok (i.e. there are no errors), False otherwise
        """
        return self.status is ErrorCode.SUCCESS

    @property
    def id(self) -> ResponseType:
        """Get the identifier of the response.

        Will vary depending on what type of response this is:

        - for a direct BLE read / write to a characteristic, it will be a :py:class:`open_gopro.constants.UUID`
        - for a BLE command response, it will be a :py:class:`open_gopro.constants.CmdType`
        - for a BLE setting / status response / update, it will be a :py:class:`open_gopro.constants.SettingId` or :py:class:`open_gopro.constants.StatusId`
        - for an HTTP response, it will be a string of the HTTP endpoint

        Returns:
            ResponseType: the identifier
        """
        return self._info[-1]

    @property
    def cmd(self) -> Optional[CmdType]:
        """Attempt to get the command ID of the response.

        If the response is not a command response, it won't have a command.

        Returns:
            Optional[CmdType]: Command ID if relevant, otherwise None
        """
        for x in self._info:
            if isinstance(x, (QueryCmdId, CmdId)):
                return x
        return None

    @property
    def uuid(self) -> Optional[UUID]:
        """Attempt to get the UUID the response was received on.

        If the response is not a BLE response, it won't have a UUID.

        Returns:
            Optional[UUID]: UUID if relevant, otherwise None
        """
        for x in self._info:
            if isinstance(x, UUID):
                return x
        return None

    @property
    def endpoint(self) -> Optional[str]:
        """Attempt to get the endpoint that response was received from.

        If the response is not an HTTP response, it won't have an endpoint.

        Returns:
            Optional[str]: Endpoint if relevant, otherwise None
        """
        for x in self._info:
            if isinstance(x, str):
                return x
        return None

    @property
    def flatten(self) -> Any:
        """Attempt to flatten / simplify the JSON response (GoProResp.data).

        - If there is only one entry in the JSON dict and it is a single value, return the value
        - If there is only one entry in the JSON dict and it is a list of values, return the list
        - Otherwise, just return the JSON dict

        Returns:
            Any: A single value, a list of values, or the JSON dict
        """
        values = list(self.data.values())
        if len(values) == 1 and type(values[0] not in [list, dict]):
            return values[0]

        if len(values) == 1 and type(values[0] is list):
            return values

        return self.data

    # to be @overload'ed if anything besides BLE notifications need to be accumulated
    def _accumulate(self, data: bytes) -> None:
        """Accumulate BLE byte data.

        If there is no more data left to accumulate, the bytestream will be parsed.

        Args:
            data (bytes): byte level BLE data
        """

        class Header(enum.Enum):
            """Packet Headers."""

            GENERAL = 0b00
            EXT_13 = 0b01
            EXT_16 = 0b10
            RESERVED = 0b11
            CONT = enum.auto()

        buf = bytearray(data)
        if buf[0] & CONT_MASK:
            buf.pop(0)
        else:
            # This is a new packet so start with an empty byte array
            self._raw_packet = bytearray([])
            hdr = Header((buf[0] & HDR_MASK) >> 5)
            if hdr is Header.GENERAL:
                self._bytes_remaining = buf[0] & GEN_LEN_MASK
                buf = buf[1:]
            elif hdr is Header.EXT_13:
                self._bytes_remaining = ((buf[0] & EXT_13_BYTE0_MASK) << 8) + buf[1]
                buf = buf[2:]
            elif hdr is Header.EXT_16:
                self._bytes_remaining = (buf[1] << 8) + buf[2]
                buf = buf[3:]

        # Append payload to buffer and update remaining / complete
        assert isinstance(self._raw_packet, bytearray)
        self._raw_packet.extend(buf)
        self._bytes_remaining -= len(buf)

        if self._bytes_remaining < 0:
            logger.error("received too much data. parsing is in unknown state")
        elif self._bytes_remaining == 0:
            self._state = GoProResp._State.ACCUMULATED

    def _parse(self) -> None:
        """Parse the accumulated response (either from a BLE bytestream or an HTTP JSON dict)."""
        # Is this a BLE response?
        if isinstance(self._raw_packet, bytearray):
            buf: bytearray = self._raw_packet

            # UUID's whose responses contain multiple parameters to parse
            if self.uuid in [UUID.CQ_QUERY_RESP, UUID.CQ_SETTINGS_RESP]:
                identifier: Optional[Type[ResponseType]] = None
                if self.uuid is UUID.CQ_SETTINGS_RESP:
                    self._info.append(SettingId(buf[0]))
                    identifier = SettingId
                else:
                    self._info.append(QueryCmdId(buf[0]))
                    if self.id in [
                        QueryCmdId.GET_SETTING_VAL,
                        QueryCmdId.REG_SETTING_VAL_UPDATE,
                        QueryCmdId.UNREG_SETTING_VAL_UPDATE,
                        QueryCmdId.SETTING_VAL_PUSH,
                        QueryCmdId.SETTING_CAPABILITY_PUSH,
                        QueryCmdId.GET_CAPABILITIES_VAL,
                        QueryCmdId.REG_CAPABILITIES_UPDATE,
                        QueryCmdId.UNREG_CAPABILITIES_UPDATE,
                    ]:
                        identifier = SettingId
                    elif self.id is QueryCmdId.GET_SETTING_NAME:
                        raise NotImplementedError
                    elif self.id in [
                        QueryCmdId.GET_STATUS_VAL,
                        QueryCmdId.REG_STATUS_VAL_UPDATE,
                        QueryCmdId.UNREG_STATUS_VAL_UPDATE,
                        QueryCmdId.STATUS_VAL_PUSH,
                    ]:
                        identifier = StatusId
                    else:
                        raise Exception("Unhandled parse state")

                # Parse all parameters
                self.status = ErrorCode(buf[1])
                buf = buf[2:]
                while len(buf) != 0:
                    param_id = identifier(buf[0])
                    param_len = buf[1]
                    buf = buf[2:]
                    # Special case where we register for a push notification for something that does not yet
                    # have a value
                    if param_len == 0:
                        self.data[param_id] = []
                        continue
                    param_val = buf[:param_len]
                    buf = buf[param_len:]

                    # Add parsed value to response's data dict
                    try:
                        # These can be more than 1 value so use a list
                        if self.cmd in [
                            QueryCmdId.GET_CAPABILITIES_VAL,
                            QueryCmdId.REG_CAPABILITIES_UPDATE,
                            QueryCmdId.SETTING_CAPABILITY_PUSH,
                        ]:
                            # Parse using parser from map and append
                            # Mypy can't follow that this parser is guarantted to be a ByteParser
                            self.data[param_id].append(self._parsers[param_id].parse(param_val))  # type: ignore
                        else:
                            # Parse using parser from map and set
                            # Mypy can't follow that this parser is guarantted to be a ByteParser
                            self.data[param_id] = self._parsers[param_id].parse(param_val)  # type: ignore
                    except KeyError:
                        # We don't have defined params for all ID's yet. Just store raw bytes
                        logger.warning(f"No parser defined for {param_id}")
                        self.data[param_id] = param_val
                    except ValueError:
                        # This is the case where we receive a value that is not defined in our params.
                        # This shouldn't happen and means the documentation needs to be updated. However, it
                        # isn't functionally critical
                        logger.warning(f"{param_id} does not contain a value {param_val}")
                        self.data[param_id] = param_val

            else:  # Other UUID's have responses that can be parsed monolithically
                if self.uuid is UUID.CQ_COMMAND_RESP:
                    self._info.append(CmdId(buf[0]))
                    # If this is a protobuf, first get the action ID (after stripping msb)
                    assert self.cmd is not None
                    if self.cmd.value >= 0xF0:
                        self._info.append(ActionId(buf[1] & 0x7F))
                        buf = buf[1:]
                    buf = buf[1:]

                try:
                    # Mypy can't follow that this parser is guarantted to be a ByteParser
                    self.data = self._parsers[self.id].parse(buf)  # type: ignore
                    self.status = self.data["status"] if "status" in self.data else self.status
                except KeyError as e:
                    self._state = GoProResp._State.ERROR
                    raise ResponseParseError(str(self.id), buf) from e

        # Is this an HTTP response?
        elif isinstance(self._raw_packet, dict):
            json_data: Dict[str, Any] = self._raw_packet
            # Is there a parser for this? Most of them do not have one yet.
            try:
                # Mypy can't follow that this parser is guaranteed to be a JsonParser
                self.data = self._parsers[self.id].parse(json_data, self._parsers)  # type: ignore
            except KeyError:
                self.data = json_data
        else:
            raise Exception(
                f"Unexpected data type ({str(type(self._raw_packet))}) when attempting to parse response"
            )

        # Recursively scrub away parsing artifacts
        scrub(self.data, "_io")
        scrub(self.data, "status")
        self._state = GoProResp._State.PARSED



# gopro.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Wed, Sep  1, 2021  5:05:47 PM

"""Implements top level interface to GoPro module."""

from __future__ import annotations
import time
import enum
import queue
import logging
import threading
from queue import Queue
from pathlib import Path
from typing import Any, Dict, Final, Optional, Type, Callable, Union, Generic, Pattern

import wrapt
import requests
from open_gopro.api.v1_0.api import ApiV1_0

from open_gopro.exceptions import InvalidOpenGoProVersion, ResponseTimeout, InvalidConfiguration
from open_gopro.ble import BLEController, UUID, BleDevice
from open_gopro.ble.adapters import BleakWrapperController
from open_gopro.wifi import WifiController
from open_gopro.wifi.adapters import Wireless
from open_gopro.util import SnapshotQueue
from open_gopro.responses import GoProResp
from open_gopro.constants import CmdId, ErrorCode, StatusId, QueryCmdId, ProducerType
from open_gopro.api import (
    Api,
    api_versions,
    BleCommands,
    BleSettings,
    BleStatuses,
    WifiCommands,
    WifiSettings,
    Params,
)
from open_gopro.communication_client import GoProBle, GoProWifi

logger = logging.getLogger(__name__)

KEEP_ALIVE_INTERVAL: Final = 60
WRITE_TIMEOUT: Final = 10
HTTP_GET_RETRIES: Final = 5


@wrapt.decorator
def _ensure_initialized_acquire_ready_semaphore(
    wrapped: Callable, instance: GoPro, args: Any, kwargs: Any
) -> Callable:
    """If the instance is initialized, acquire the semaphore before doing anything.

    Raises:
        GoProNotInitialized: The function can't be used yet because the GoPro device isn't initialized

    Returns:
        Callable: Function to call after semaphore has been acquired
    """
    if instance._maintain_ble:
        logger.debug(f"{wrapped.__name__} acquiring semaphore")
        with instance._ready:
            logger.debug(f"{wrapped.__name__} has the semaphore")
            ret = wrapped(*args, **kwargs)
    else:
        ret = wrapped(*args, **kwargs)
    if instance._maintain_ble:
        logger.debug(f"{wrapped.__name__} released the semaphore")
    return ret


class GoPro(GoProBle, GoProWifi, Generic[BleDevice]):
    """The top-level BLE and Wifi interface to a GoPro device.

    See `Open GoPro <https://gopro.github.io/OpenGoPro/python_sdk>`_ for complete documentation.

    This will handle for BLE:

    - discovering device
    - establishing connections
    - discovering GATT characteristics
    - enabling notifications
    - discovering Open GoPro version
    - transferring data

    This will handle for Wifi:

    - finding SSID and password
    - establishing Wifi connection
    - transferring data

    It will also do some synchronization, etc:

    - ensuring camera is ready / not encoding before transferring data
    - sending keep alive signal periodically

    If no target arg is passed in, the first discovered GoPro device will be connected to.

    It can be used via context manager:

    >>> with GoPro() as gopro:
    >>>     gopro.ble_command.set_shutter(gopro.params.Shutter.ON)

    Or without:

    >>> gopro = GoPro()
    >>> gopro.open()
    >>> gopro.ble_command.set_shutter(gopro.params.Shutter.ON)
    >>> gopro.close()

    Args:
        identifier (Pattern, optional): Last 4 of camera name / serial number (i.e. 0456 for GoPro0456). Defaults
            to None (i.e. connect to first discovered GoPro)
        ble_adapter (BLEController, optional): Class used to control computer's BLE connection / send data.
            Defaults to BleakController().
        wifi_adapter (WifiController, optional): Class used to control computer's Wifi connection / send data.
            Defaults to Wireless().
        enable_wifi (bool, optional): Optionally do not enable Wifi if set to False. Defaults to True.
        maintain_ble (bool, optional): Optionally do not perform BLE housekeeping if set to False (used for
            testing). Defaults to True.
    """

    _base_url = "http://10.5.5.9:8080/"  #: Hard-coded Open GoPro base URL

    class _InternalState(enum.IntFlag):
        """State used to manage whether the GoPro instance is ready or not."""

        READY = 0
        ENCODING = 1 << 0
        SYSTEM_BUSY = 1 << 1

    def __init__(
        self,
        target: Optional[Union[Pattern, BleDevice]] = None,
        ble_adapter: Type[BLEController] = BleakWrapperController,
        wifi_adapter: Type[WifiController] = Wireless,
        enable_wifi: bool = True,
        maintain_ble: bool = True,
    ) -> None:
        # Store initialization information
        self._enable_wifi_during_init = enable_wifi
        self._maintain_ble = maintain_ble

        # Initialize GoPro Communication Client
        GoProBle.__init__(self, ble_adapter(), self._disconnect_handler, self._notification_handler, target)
        GoProWifi.__init__(self, wifi_adapter())

        # We start with version 1.0. It will be updated once we query the version
        self._api: Api = ApiV1_0(self, self)

        # Current accumulating synchronous responses, indexed by UUID. This assumes there can only be one active response per UUID
        self._active_resp: Dict[UUID, GoProResp] = {}
        # Responses that we are waiting for.
        self._sync_resp_wait_q: SnapshotQueue = SnapshotQueue()
        # Synchronous response that has been parsed and are ready for their sender to receive as the response.
        self._sync_resp_ready_q: SnapshotQueue = SnapshotQueue()

        # For outputting asynchronously received information
        self._out_q: "Queue[GoProResp]" = Queue()
        self._listeners: Dict[ProducerType, bool] = {}

        # Set up events
        self._ble_disconnect_event = threading.Event()
        self._ble_disconnect_event.set()

        # Set up threads
        self._threads_waiting = 0
        # If we are to perform BLE housekeeping
        if self._maintain_ble:
            self._threads_waiting += 2
            # Set up thread to send keep alive
            self._keep_alive_thread = threading.Thread(
                target=self._periodic_keep_alive, name="keep_alive", daemon=True
            )
            # Set up thread to block until camera is ready to receive commands
            self._ready = threading.BoundedSemaphore(value=1)
            self._state_condition = threading.Condition()
            self._internal_state = GoPro._InternalState.ENCODING | GoPro._InternalState.SYSTEM_BUSY
            self._state_thread = threading.Thread(target=self._maintain_state, name="state", daemon=True)

    def __enter__(self) -> "GoPro":  # pylint: disable=missing-return-doc
        self.open()
        return self

    def __exit__(self, *_: Any) -> None:
        self.close()

    def __del__(self) -> None:
        self.close()

    @property
    def identifier(self) -> Optional[str]:
        """Get a unique identifier for this instance.

        The identifier is the last 4 digits of the camera. That is, the same string that is used to
        scan for the camera for BLE.

        If no target has been provided and a camera is not yet found, this will be None

        Returns:
            Optional[str]: last 4 digits if available, else None
        """
        return self._ble.identifier

    @property
    def is_ble_connected(self) -> bool:
        """Are we connected via BLE to the GoPro device?

        Returns:
            bool: True if yes, False if no
        """
        return self._ble.is_connected

    @property
    def is_wifi_connected(self) -> bool:
        """Are we connected via Wifi to the GoPro device?

        Returns:
            bool: True if yes, False if no
        """
        return self._wifi.is_connected

    @property
    def is_encoding(self) -> bool:
        """Is the camera currently encoding?

        Returns:
            bool: True if yes, False if no
        """
        if not self._maintain_ble:
            raise InvalidConfiguration("Not maintaining BLE state so encoding is not applicable")
        return self._internal_state & GoPro._InternalState.ENCODING == 1

    @property
    def is_busy(self) -> bool:
        """Is the camera currently performing a task that prevents it from accepting commands?

        Returns:
            bool: True if yes, False if no
        """
        if not self._maintain_ble:
            raise InvalidConfiguration("Not maintaining BLE state so busy is not applicable")
        return self._internal_state & GoPro._InternalState.SYSTEM_BUSY == 1

    @property
    def version(self) -> float:
        """The API version does the connected camera supports

        Note! If we have not yet connected and query the peer to find its version, this will be set to 1.0

        Returns:
            float: supported version in decimal form
        """
        return float(self._api.version)

    @property
    def ble_command(self) -> BleCommands:
        """Used to call the version-specific BLE commands

        Returns:
            BleCommands: the commands
        """
        return self._api.ble_command

    @property
    def ble_setting(self) -> BleSettings:
        """Used to access the version-specific BLE settings

        Returns:
            BleSettings: the settings
        """
        return self._api.ble_setting

    @property
    def ble_status(self) -> BleStatuses:
        """Used to access the version-specific BLE statuses

        Returns:
            BleStatuses: the statuses
        """
        return self._api.ble_status

    @property
    def wifi_command(self) -> WifiCommands:
        """Used to access the version-specific Wifi commands

        Returns:
            WifiCommands: the commands
        """
        return self._api.wifi_command

    @property
    def wifi_setting(self) -> WifiSettings:
        """Used to access the version-specific Wifi settings

        Returns:
            WifiSettings: the settings
        """
        return self._api.wifi_setting

    @property
    def params(self) -> Type[Params]:
        """Version-specific parameters for BLE / Wifi commands, statuses, and settings

        Returns:
            Type[Params]: the parameters
        """
        return self._api.params

    def open(self, timeout: int = 10, retries: int = 5) -> None:
        """Perform all initialization commands for ble and wifi

        For BLE: scan and find device, establish connection, discover characteristics, configure queries
        start maintenance, and get Open GoPro version..

        For Wifi: discover SSID and password, enable and connect. Or disable if not using.

        Args:
            timeout (int, optional): How long to wait for each connection before timing out. Defaults to 10.
            retries (int, optional): How many connection attempts before considering connection failed. Defaults to 5.
        """
        # Establish BLE connection and start maintenance threads if desired
        self._open_ble(timeout, retries)

        # Find and configure API version
        version = self.ble_command.get_open_gopro_api_version().flatten
        version_str = f"{version.major}.{version.minor}"
        try:
            self._api = api_versions[version_str](self, self)
        except KeyError as e:
            raise InvalidOpenGoProVersion(version_str) from e
        logger.info(f"Using Open GoPro API version {version_str}")

        # Establish Wifi connection if desired
        if self._enable_wifi_during_init:
            self._open_wifi(timeout, retries)
        else:
            # Otherwise, turn off Wifi
            logger.info("Turning off the camera's Wifi radio")
            self.ble_command.enable_wifi_ap(False)

    def close(self) -> None:
        """Safely stop the GoPro instance.

        This will disconnect BLE and WiFI if applicable.

        If not using the context manager, it is mandatory to call this before exiting the program in order to
        prevent reconnection issues because the OS has never disconnected from the previous session.
        """
        self._close_wifi()
        self._close_ble()

    def get_update(self, timeout: float = None) -> GoProResp:
        """Get a notification that we received from a registered listener.

        If timeout is None, this will block until a notification is received.
        The updates are received via FIFO

        Args:
            timeout (float, optional): Time to wait for a notification before returning. Defaults to None (wait forever)

        Returns:
            GoProResp: Received notification
        """
        return self._out_q.get(timeout=timeout)

    def keep_alive(self) -> bool:
        """Send a heartbeat to prevent the BLE connection from dropping.

        This is sent automatically by the GoPro instance if its `maintain_ble` argument is not False.

        Returns:
            bool: True if it succeeded,. False otherwise
        """
        return self.ble_setting.led.set(self.params.LED.BLE_KEEP_ALIVE).is_ok

    ##########################################################################################################
    #                                 End Public API
    ##########################################################################################################

    @property
    def _is_ble_initialized(self) -> bool:
        """Are we done

        Returns:
            bool: True if yes, False if no
        """
        return self._threads_waiting == 0

    def _maintain_state(self) -> None:
        """Thread to keep track of ready / encoding and acquire / release ready semaphore."""
        self._ready.acquire()
        while self.is_ble_connected:
            internal_status_previous = self._internal_state
            with self._state_condition:
                self._state_condition.wait()
                # If we were ready but not now we're not, acquire the semaphore
                if internal_status_previous == 0 and self._internal_state != 0:
                    logger.debug("Control acquiring semaphore")
                    self._ready.acquire()
                    logger.debug("Control has semaphore")
                # If we weren't ready but now we are, release the semaphore
                elif internal_status_previous != 0 and self._internal_state == 0:
                    # If this is the first time, mark that we might now be initialized
                    if not self._is_ble_initialized:
                        self._threads_waiting -= 1
                    self._ready.release()
                    logger.debug("Control released semaphore")

        self._threads_waiting += 1
        logger.debug("Maintain state thread exiting...")

    def _periodic_keep_alive(self) -> None:
        """Thread to periodically send the keep alive message via BLE."""
        while self.is_ble_connected:
            if not self._is_ble_initialized:
                self._threads_waiting -= 1
            try:
                if self.keep_alive():
                    time.sleep(KEEP_ALIVE_INTERVAL)
            except Exception:  # pylint: disable=broad-except
                # If the connection disconnects while we were trying to send, there can be any number
                # of exceptions. This is expected and this thread will exit on the next while check.
                pass

        self._threads_waiting += 1
        logger.debug("periodic keep alive thread exiting...")

    def _register_listener(self, producer: ProducerType) -> None:
        """Register a producer to store notifications from.

        The notifications can be accessed via the get_update() method.

        Args:
            producer (ProducerType): Producer to listen to.
        """
        self._listeners[producer] = True

    def _unregister_listener(self, producer: ProducerType) -> None:
        """Unregister a producer in order to stop listening to its notifications.

        Args:
            producer (ProducerType): Producer to stop listening to.
        """
        if producer in self._listeners:
            del self._listeners[producer]

    def _open_ble(self, timeout: int = 10, retries: int = 5) -> None:
        """Connect the instance to a device via BLE.

        Args:
            device (BleDevice): Device to connect to
            timeout (int, optional): Time in seconds before considering establishment failed. Defaults to 10 seconds.
            retries (int, optional): How many tries to reconnect after failures. Defaults to 5.

        Raises:
            ConnectFailed: Connection could not be established
        """
        # Establish connection, pair, etc.
        self._ble.open(timeout, retries)
        # Configure threads if desired
        if self._maintain_ble:
            self._state_thread.start()
            self.ble_status.encoding_active.register_value_update()
            self.ble_status.system_ready.register_value_update()
            self._keep_alive_thread.start()
        logger.info("BLE is ready!")

    # TODO refactor this into smaller methods
    def _notification_handler(self, handle: int, data: bytearray) -> None:
        """Receive notifications from the BLE controller.

        Args:
            handle (int): Attribute handle that notification was received on.
            data (bytes): Bytestream that was received.
        """
        # Convert handle to UUID
        uuid = self._ble.gatt_table.handle2uuid(handle)
        # Responses we don't care about. For now, just the BLE-spec defined battery characteristic
        if uuid is UUID.BATT_LEVEL:
            return
        logger.debug(f'Received response on {uuid}: {data.hex(":")}')

        # Add to response dict if not already there
        if uuid not in self._active_resp:
            self._active_resp[uuid] = GoProResp(self._parser_map, info=[uuid])

        self._active_resp[uuid]._accumulate(data)

        if self._active_resp[uuid].is_received:
            response = self._active_resp[uuid]
            response._parse()

            # Handle internal statuses
            if self._maintain_ble:
                if (
                    response.cmd
                    in [
                        QueryCmdId.REG_STATUS_VAL_UPDATE,
                        QueryCmdId.GET_STATUS_VAL,
                        QueryCmdId.STATUS_VAL_PUSH,
                    ]
                    and StatusId.ENCODING in response.data
                ):
                    with self._state_condition:
                        if response[StatusId.ENCODING] is True:
                            self._internal_state |= GoPro._InternalState.ENCODING
                        else:
                            self._internal_state &= ~GoPro._InternalState.ENCODING
                        self._state_condition.notify()
                if (
                    response.cmd
                    in [
                        QueryCmdId.REG_STATUS_VAL_UPDATE,
                        QueryCmdId.GET_STATUS_VAL,
                        QueryCmdId.STATUS_VAL_PUSH,
                    ]
                    and StatusId.SYSTEM_READY in response.data
                ):
                    with self._state_condition:
                        if response[StatusId.SYSTEM_READY] is True:
                            self._internal_state &= ~GoPro._InternalState.SYSTEM_BUSY
                        else:
                            self._internal_state |= GoPro._InternalState.SYSTEM_BUSY
                        self._state_condition.notify()

            # Check if this is the awaited synchronous response (id matches). Note! these have to come in order.
            response_claimed = False
            if not self._sync_resp_wait_q.empty():
                queue_snapshot = self._sync_resp_wait_q.snapshot()
                if queue_snapshot[0].id is response.id:
                    # Dequeue it and put this on the ready queue
                    self._sync_resp_wait_q.get_nowait()
                    self._sync_resp_ready_q.put(response)
                    response_claimed = True

            # If this wasn't the awaited synchronous response...
            if not response_claimed:
                logger.info(f"--(ASYNC)--> {response}")
                # See if there are any registered responses that need to be enqueued for client consumption
                for key in list(response.data.keys()):
                    if (response.cmd, key) not in self._listeners:
                        del response.data[key]
                # Enqueue the response if there is anything left
                if len(response.data) > 0:
                    self._out_q.put(response)

            # Clear active response from response dict
            del self._active_resp[uuid]

    def _close_ble(self) -> None:
        if self.is_ble_connected and self._ble is not None:
            logger.info("Terminating the BLE connection")
            self._ble_disconnect_event.clear()
            self._ble.close()
            self._ble_disconnect_event.wait()

    def _disconnect_handler(self, _: Any) -> None:
        """Handle disconnects"""
        if self._ble_disconnect_event.is_set():
            logger.error("Ble connection terminated unexpectedly.")
            # TODO how to handle unexpected disconnects?
            # raise ConnectionTerminated("Ble connection terminated.")
        self._ble_disconnect_event.set()

    def _write_characteristic_receive_notification(self, uuid: UUID, data: bytearray) -> GoProResp:
        """Perform a BLE write and wait for a corresponding notification response.

        There should hopefully not be a scenario where this needs to be called directly as it is generally
        called from the instance's API delegate (i.e. self)

        Args:
            uuid (UUID): UUID to write to
            data (bytearray): data to write

        Raises:
            Exception: Unexpected functionality occurred

        Returns:
            GoProResp: parsed notification response data
        """
        assert self._ble is not None
        # Acquire ready semaphore unless we are initializing or this is a Set Shutter Off command
        have_semaphore = False
        if (
            self._maintain_ble
            and self._is_ble_initialized
            and not (
                GoProResp._from_write_command(self._parser_map, uuid, data).id is CmdId.SET_SHUTTER
                and data[-1] == 0
            )
        ):
            logger.debug(
                f"{GoProResp._from_write_command(self._parser_map, uuid, data).id} acquiring semaphore"
            )
            self._ready.acquire()
            logger.debug(f"{GoProResp._from_write_command(self._parser_map, uuid, data).id} has semaphore")
            have_semaphore = True

        # Store information on the response we are expecting
        self._sync_resp_wait_q.put(GoProResp._from_write_command(self._parser_map, uuid, data))
        # Perform write
        self._ble.write(uuid.value, data)
        # Wait to be notified that response was received
        try:
            response = self._sync_resp_ready_q.get(timeout=WRITE_TIMEOUT)
        except queue.Empty as e:
            logger.error(f"Response timeout of {WRITE_TIMEOUT} seconds!")
            raise ResponseTimeout(WRITE_TIMEOUT) from e

        # Check status
        try:
            if response.status is not ErrorCode.SUCCESS:
                logger.warning(f"Received non-success status: {response.status}")
        except AttributeError:
            logger.error("Not able to parse status from response")

        if self._maintain_ble:
            # If this was set shutter on, we need to wait to be notified that encoding has started
            if response.cmd is CmdId.SET_SHUTTER and data[-1] == 1:
                while not self.is_encoding:
                    # We don't want to use the application's loop, can't use any of our loops due to potential deadlock,
                    # and don't want to spawn a new thread for this. So just poll ¯\_(ツ)_/¯
                    # A read to an int is atomic anyway.
                    time.sleep(0.1)
            # Release the semaphore if we acquired it
            if have_semaphore:
                self._ready.release()
                logger.debug(
                    f"{GoProResp._from_write_command(self._parser_map, uuid, data).id} released the semaphore"
                )

        return response

    def _read_characteristic(self, uuid: UUID) -> GoProResp:
        """Read a characteristic's data by UUID.

        There should hopefully not be a scenario where this needs to be called directly as it is generally
        called from the instance's delegates (i.e. self.command, self.setting, self.ble_status)

        Args:
            uuid (UUID): characteristic data to read

        Returns:
            bytearray: read data
        """
        assert self._ble is not None
        have_semaphore = False
        if self._maintain_ble:
            logger.debug(f"{uuid} acquiring semaphore")
            self._ready.acquire()
            logger.debug(f"{uuid} has the semaphore")
            have_semaphore = True

        received_data = self._ble.read(uuid.value)

        if self._maintain_ble and have_semaphore:
            self._ready.release()
            logger.debug(f"{uuid} released the semaphore")

        return GoProResp._from_read_response(self._parser_map, uuid, received_data)

    def _open_wifi(self, timeout: int = 15, retries: int = 5) -> None:
        """Connect to a GoPro device via Wifi.

        Args:
            enable (bool): whether to enable or disable wifi
            timeout (int, optional): Time before considering establishment failed. Defaults to 15 seconds.
            retries (int, optional): How many tries to reconnect after failures. Defaults to 5.

        Raises:
            Exception: Wifi failed to connect.
        """
        logger.info("Discovering Wifi AP info and enabling via BLE")
        password = self.ble_command.get_wifi_password().flatten
        ssid = self.ble_command.get_wifi_ssid().flatten
        self.ble_command.enable_wifi_ap(True)
        self._wifi.open(ssid, password, timeout, retries)

    def _close_wifi(self) -> None:
        """Terminate the Wifi connection."""
        self._wifi.close()

    @_ensure_initialized_acquire_ready_semaphore
    def _get(self, url: str) -> GoProResp:
        """Send an HTTP GET request to an Open GoPro endpoint.

        There should hopefully not be a scenario where this needs to be called directly as it is generally
        called from the instance's delegates (i.e. self.wifi_command and self.wifi_status)

        Args:
            url (str): endpoint URL

        Returns:
            GoProResp: response
        """
        url = GoPro._base_url + url
        logger.debug(f"Sending:  {url}")

        response: Optional[GoProResp] = None
        for retry in range(HTTP_GET_RETRIES):
            try:
                request = requests.get(url)
                request.raise_for_status()
                response = GoProResp._from_http_response(self._parser_map, request)
                break
            except requests.exceptions.HTTPError as e:
                # The camera responded with an error. Break since we successfully sent the command and attempt
                # to continue
                logger.warning(e)
                response = GoProResp._from_http_response(self._parser_map, e.response)
                break
            # TODO figure out why these are failing. For now just retry
            except requests.exceptions.ConnectionError as e:
                logger.warning(repr(e))
                logger.warning("Retrying to send the command...")
                if retry == HTTP_GET_RETRIES - 1:
                    raise ResponseTimeout(HTTP_GET_RETRIES) from e

        assert response is not None
        return response

    @_ensure_initialized_acquire_ready_semaphore
    def _stream_to_file(self, url: str, file: Path) -> None:
        """Send an HTTP GET request to an Open GoPro endpoint to download a binary file.

        There should hopefully not be a scenario where this needs to be called directly as it is generally
        called from the instance's delegates (i.e. self.wifi_command and self.wifi_status)

        Args:
            url (str): endpoint URL
            file (Path): location where file should be downloaded to
        """
        url = GoPro._base_url + url
        logger.debug(f"Sending: {url}")
        with requests.get(url, stream=True) as request:
            request.raise_for_status()
            with open(file, "wb") as f:
                logger.debug(f"receiving stream to {file}...")
                for chunk in request.iter_content(chunk_size=8192):
                    f.write(chunk)



# __version__.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Wed, Sep  1, 2021  5:05:51 PM

"""Single source of versioning for Open GoPro Python SDK"""

__version__ = "0.7.1"



# set_camera_control_status_pb.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Wed, Nov 17, 2021  9:12:09 PM

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: set_camera_control_status.proto
# plugin: python-betterproto
from dataclasses import dataclass

import betterproto


class EnumCameraControlStatus(betterproto.Enum):
    CAMERA_IDLE = 0
    CAMERA_CONTROL = 1
    CAMERA_EXTERNAL_CONTROL = 2


@dataclass
class RequestSetCameraControlStatus(betterproto.Message):
    camera_control_status: "EnumCameraControlStatus" = betterproto.enum_field(1)



# __init__.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Wed, Sep  1, 2021  5:05:51 PM

"""All GoPro exports that the the user will want should be exported here."""

from open_gopro.gopro import GoPro
from open_gopro.__version__ import __version__
from open_gopro.responses import GoProResp



# api.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue Sep  7 21:35:53 UTC 2021

"""Implementation of Open GoPro API version 2.0"""

from typing import Type

from open_gopro.api.v1_0.api import ApiV1_0
from open_gopro.communication_client import GoProBle, GoProWifi
from .ble_commands import BleCommandsV2_0, BleSettingsV2_0, BleStatusesV2_0
from .wifi_commands import WifiCommandsV2_0, WifiSettingsV2_0
from .params import ParamsV2_0


class ApiV2_0(ApiV1_0):
    """Implementation of Open GoPro API version 2.0"""

    version = "2.0"

    def __init__(  # pylint: disable = super-init-not-called
        self, ble_communicator: GoProBle, wifi_communicator: GoProWifi
    ) -> None:  # pylint: disable = super-init-not-called
        # No call to superclass's __init__ since we are overriding each individual attribute below. Would be redundant.
        self.params: Type[ParamsV2_0] = ParamsV2_0
        self.ble_command: BleCommandsV2_0 = BleCommandsV2_0(ble_communicator)
        self.ble_setting: BleSettingsV2_0 = BleSettingsV2_0(ble_communicator, self.params)
        self.ble_status: BleStatusesV2_0 = BleStatusesV2_0(ble_communicator, self.params)
        self.wifi_command: WifiCommandsV2_0 = WifiCommandsV2_0(wifi_communicator)
        self.wifi_setting: WifiSettingsV2_0 = WifiSettingsV2_0(wifi_communicator, self.params)



# ble_commands.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue Sep  7 21:35:53 UTC 2021

"""Updates to BLE API for Open GoPro version 2.0"""

from __future__ import annotations
import logging
from typing import Type

from construct import Flag, Int8ub

from open_gopro.constants import StatusId, SettingId
from open_gopro.api.v1_0.ble_commands import BleCommandsV1_0, BleSettingsV1_0, BleStatusesV1_0
from open_gopro.communication_client import GoProBle
from open_gopro.api.builders import BleSetting, BleStatus, build_enum_adapter
from .params import ParamsV2_0 as Params

logger = logging.getLogger(__name__)


class BleCommandsV2_0(BleCommandsV1_0):
    """Implement updates to BLE commands for version 2.0"""


class BleSettingsV2_0(BleSettingsV1_0):
    # pylint: disable=missing-class-docstring, unused-argument
    """Implement updates to BLE Settings for version 2.0"""

    def __init__(self, communicator: GoProBle, params: Type[Params]) -> None:
        super().__init__(communicator, params)

        self.video_performance_mode = BleSetting[Params.VideoPerformanceMode](
            self.communicator,
            SettingId.VIDEO_PERFORMANCE_MODE,
            build_enum_adapter(params.VideoPerformanceMode),
        )
        """Video Performance Mode. Set with :py:class:`Params.VideoPerformanceMode`"""


class BleStatusesV2_0(BleStatusesV1_0):
    """Implement updates to BLE statuses for version 2.0"""

    def __init__(self, communicator: GoProBle, params: Type[Params]) -> None:
        super().__init__(communicator, params)

        self.sd_rating_check_error: BleStatus = BleStatus(
            self.communicator, StatusId.SD_RATING_CHECK_ERROR, Flag
        )
        """Does sdcard meet specified minimum write speed?"""

        self.sd_write_speed_error: BleStatus = BleStatus(
            self.communicator, StatusId.SD_WRITE_SPEED_ERROR, Int8ub
        )
        """Number of sdcard write speed errors since device booted"""

        self.camera_control: BleStatus = BleStatus(
            self.communicator, StatusId.CAMERA_CONTROL, build_enum_adapter(params.CameraControl)
        )
        """Camera control status ID"""

        self.usb_connected: BleStatus = BleStatus(self.communicator, StatusId.USB_CONNECTED, Flag)
        """Is the camera connected to a PC via USB?"""



# api.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue Sep  7 21:35:52 UTC 2021

"""Implementation of Open GoPro API version 1.0"""

from open_gopro.communication_client import GoProBle, GoProWifi
from .ble_commands import BleCommandsV1_0, BleSettingsV1_0, BleStatusesV1_0
from .wifi_commands import WifiCommandsV1_0, WifiSettingsV1_0
from .params import ParamsV1_0


class ApiV1_0:
    """Implementation of Open GoPro API version 1.0"""

    version = "1.0"

    def __init__(self, ble_communicator: GoProBle, wifi_communicator: GoProWifi) -> None:
        self.params = ParamsV1_0
        self.ble_command = BleCommandsV1_0(ble_communicator)
        self.ble_setting = BleSettingsV1_0(ble_communicator, self.params)
        self.ble_status = BleStatusesV1_0(ble_communicator, self.params)
        self.wifi_command = WifiCommandsV1_0(wifi_communicator)
        self.wifi_setting = WifiSettingsV1_0(wifi_communicator, self.params)
